# -*- coding:utf-8 -*-

# Copyright © 2017-2019, Institut Pasteur
#    Contributor: Maxime Duval

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

"""
This module regroups random walk simulations of many different kinds.
"""

import numpy as np
import numpy.matlib as npm
import pandas as pd

from .distribution import *
from .rw_misc import *


# Directory of saved fractal patterns (to avoid generating them each time
# we want a random walk).
FRACTAL_DIR = ('Z:\\LAB_shared_stuff\\Maxime_Duval\\single_rw\\'
               'Fractal_pattern\\patterns')
TREE_DIR = ('Z:\\LAB_shared_stuff\\Maxime_Duval\\single_rw\\DLA\\patterns')


# Continous time random walks

def RW_brownian(T_max=1, dt=1e-2, D=0.1, nb_short=1, v=None, X_init=0, dim=2):
    """
    Generates simple pure diffusive random walk.

    Parameters
    ----------
    T_max : float, the total time computed for which the random walk is
        computed.
    dt : float, time between each point output.
    D : float, the diffusivity of the random walk.
    nb_short : int,  the number of points used between each point output, to
        avoid accumulating noise.
    X_init : float or list, initial position of the random walk.
    dim : int, in {1,2,3}, the dimension of the random walk.
    """
    nb_point = int(T_max/dt)
    nb_tot = (nb_point-1) * nb_short
    dt_short = dt/nb_short
    Xi = np.random.randn(nb_tot, dim)
    X_init = normalize_init(X_init, dim)
    drift = np.array(normalize_init(v, dim)) * dt * np.ones((nb_point-1, dim))
    X = np.cumsum(np.sqrt(2*D*dt_short)*Xi, axis=0) + np.array(X_init) + drift
    X = X[::nb_short]
    X = np.insert(X, 0, X_init, axis=0)
    t = np.arange(nb_point)*dt
    data = np.stack((t,) + tuple(X[:, i] for i in range(dim))).T
    return pd.DataFrame(data=data, columns=['t'] + SPACE_COLS[:dim])


def RW_exp_dist(T_max=1, dt=1e-2, d_l=0.1, v=None, X_init=0, dim=2):
    return CTRW(T_max=T_max, dt=dt, distribution_space="exp", d_l=d_l,
                distribution_time="cst", v=v, X_init=X_init, dim=dim)


def RW_const_dist(T_max=1, dt=1e-2, d_l=0.1, v=None, X_init=0, dim=2):
    return CTRW(T_max=T_max, dt=dt, distribution_space="uni", d_l=d_l,
                distribution_time="cst", v=v, X_init=X_init, dim=dim)


def RW_gauss_dist(T_max=1, dt=1e-2, d_l=0.1, v=None, X_init=0, dim=2):
    return CTRW(T_max=T_max, dt=dt, distribution_space="gauss", d_l=d_l,
                distribution_time="cst", v=v, X_init=X_init, dim=dim)


def RW_exp_dist_exp_temps(T_max=1, dt=1e-2, d_l=0.1,
                          d_tau=1e-2, v=None, X_init=0, dim=2):
    return CTRW(T_max=T_max, dt=dt, distribution_space="exp", d_l=d_l,
                distribution_time="exp", d_tau=d_tau, v=v, X_init=X_init,
                dim=dim)


def RW_anomalous_dist(T_max=1, dt=1e-2, alpha=1.5, d_scale=0.1,
                      c_scale_alpha_stable=1, nature_distribution="lomax",
                      v=None, X_init=0, dim=2):
    return CTRW(T_max=T_max, dt=dt, distribution_space=nature_distribution,
                c_scale_alpha_stable_x=c_scale_alpha_stable,
                d_l=d_scale, alpha_space=alpha,
                distribution_time="cst", v=v, X_init=X_init, dim=dim)


def RW_anomalous_with_cut_dists(T_max=1, dt=1e-2, alpha=1.5, d_scale=0.1,
                                d_jump_max=1, v=None,
                                nature_distribution="constant_tail_cut",
                                X_init=0, dim=2):
    return CTRW(T_max=T_max, dt=dt, distribution_space=nature_distribution,
                d_l=d_scale, alpha_space=alpha, d_jump_max=d_jump_max,
                distribution_time="cst", v=v, X_init=X_init, dim=dim)


def RW_exp_dist_anomalous_time(T_max=1, dt=1e-2, d_l=0.1, alpha=1.5, d_tau=0.1,
                               c_scale_alpha_stable=1, v=None,
                               nature_distribution="lomax", X_init=0, dim=2):
    return CTRW(T_max=T_max, dt=dt, distribution_space="exp", d_l=d_l,
                distribution_time=nature_distribution, alpha_time=alpha,
                d_tau=d_tau, v=v, X_init=X_init, dim=dim,
                c_scale_alpha_stable_t=c_scale_alpha_stable)


def RW_exp_dist_anomalous_with_cut_time(T_max=1, dt=1e-2, d_l=0.1, alpha=1.5,
                                        d_tau=0.1, d_tau_max=1,
                                        nature_distribution="lomax_cut",
                                        v=None, X_init=0, dim=2):
    return CTRW(T_max=T_max, dt=dt, distribution_space="exp", d_l=d_l,
                distribution_time=nature_distribution, alpha_time=alpha,
                d_tau=d_tau, d_tau_max=d_tau_max, v=v, X_init=X_init, dim=dim)


def RW_gauss_dist_anomalous_with_cut_time(T_max=1, dt=1e-2, d_l=0.1, alpha=1.5,
                                          d_tau=0.1, d_tau_max=1,
                                          nature_distribution="lomax_cut",
                                          v=None, X_init=0, dim=2):
    return CTRW(T_max=T_max, dt=dt, distribution_space="gauss", d_l=d_l,
                distribution_time=nature_distribution, alpha_time=alpha,
                d_tau=d_tau, d_tau_max=d_tau_max, v=v, X_init=X_init, dim=dim)


def RW_const_dist_anomalous_time(T_max=1, dt=1e-2, d_l=0.1, alpha=1.5,
                                 d_tau=0.1, d_tau_max=1,
                                 c_scale_alpha_stable=1,
                                 nature_distribution="lomax",
                                 v=None, X_init=0, dim=2):
    return CTRW(T_max=T_max, dt=dt, distribution_space="uni", d_l=d_l,
                distribution_time=nature_distribution, alpha_time=alpha,
                d_tau=d_tau, d_tau_max=d_tau_max, v=v, X_init=X_init, dim=dim,
                c_scale_alpha_stable_t=c_scale_alpha_stable)


def RW_const_dist_anomalous_with_cut_time(T_max=1, dt=1e-2, d_l=0.1, alpha=1.5,
                                          d_tau=0.1, d_tau_max=1,
                                          nature_distribution="lomax_cut",
                                          v=None, X_init=0, dim=2):
    return CTRW(T_max=T_max, dt=dt, distribution_space="uni", d_l=d_l,
                distribution_time=nature_distribution, alpha_time=alpha,
                d_tau=d_tau, d_tau_max=d_tau_max, v=v, X_init=X_init, dim=dim)


def CTRW(T_max=1, dt=1e-2,
         distribution_space="exp",  alpha_space=1.5, d_l=0.1, d_jump_max=1,
         c_scale_alpha_stable_x=1,
         distribution_time="cst", alpha_time=1.5, d_tau=0.02, d_tau_max=1,
         c_scale_alpha_stable_t=1,
         v=None, X_init=0, dim=2):
    """
    Generates diffusive RW from chosen distribution of distances and waiting
    times.

    Parameters
    ----------
    T_max : float, the total time computed for which the random walk is
        computed.
    dt : float, time between each point output.
    distribution_space : string, the type of the distance distribution.
    alpha_space : float, parameter used for long tail distributions.
    d_l : float, the characteristic distance of the random walk.
    d_jump_max : float, the maximum distance of a jump for long tail
        distributions with a cut.
    c_scale_alpha_stable_x : scale parameter in case distances follow an alpha
        levy stable distribution.
    distribution_time : string, the type of the waiting time distribution.
    alpha_time : float, parameter used for long tail distributions.
    d_tau : float, the characteristic waiting time of the random walk.
    d_tau_max : float, the maximum distance of a jump for long tail
        distributions with a cut.
    c_scale_alpha_stable_x : scale parameter in case waiting times follow an
        alpha levy stable distribution.
    v : drift, list of size dim.
    X_init : float or list, initial position of the random walk.
    dim : int, in {1,2,3}, the dimension of the random walk.
    """
    nb_point = int(T_max/dt)
    if distribution_time != "cst":
        nb_point *= 10
        T_max *= 10
    # Generates a vector t such that t[-1] > T ==> len(t) >= nb_point
    t = generate_times(distribution_time, nb_point, T_max, dt,
                       alpha=alpha_time,
                       d_tau=d_tau, d_tau_max=d_tau_max, dim=dim,
                       c_scale_alpha_stable_t=c_scale_alpha_stable_t)
    raw_length = len(t)
    distances = generate_distribution(distribution_space,
                                      raw_length - 1, alpha=alpha_space,
                                      d_l=d_l, d_max=d_jump_max, dim=dim,
                                      c_alpha_levy=c_scale_alpha_stable_x)
    random_angles = distrib_random_angle(raw_length-1, dim=dim)
    X = apply_angle_dists(distances, random_angles, dim)
    normed_pos_init = np.array(normalize_init(X_init, dim))
    drift = np.array(normalize_init(v, dim)) * dt * np.ones(X.shape)
    X = X.cumsum(axis=0) + normed_pos_init + drift
    X = np.concatenate((np.expand_dims(normed_pos_init, axis=0), X), axis=0)
    if distribution_time != "cst":
        t_regularized = np.linspace(0, T_max, nb_point, endpoint=False)
        X, t = regularize_times(X, t, t_regularized)
    data = np.concatenate((np.expand_dims(t, axis=1), X), axis=1)
    if distribution_time != "cst":
        random_start = np.random.randint(0, int(nb_point * 0.9))
        data = data[random_start:random_start+int(nb_point * 0.1)]
        data[:, 0] = np.linspace(0, T_max*0.1, int(nb_point*0.1),
                                 endpoint=False)
        data[:, [1, 2]] -= data[0, [1, 2]]
    return pd.DataFrame(data=data, columns=['t'] + SPACE_COLS[:dim])

# Fractional random walks.


def RW_FBM(T_max=1, dt=1e-2, corr_type="frac", sigma=0.5, H=0.55,
           v=None, X_init=0, dim=2):
    """
    Generates fractal diffusive RW.

    Parameters
    ----------
    T_max : float, the total time computed for which the random walk is
        computed.
    dt : float, time between each point output.
    corr_type : string, type of correlation. Should be in {"frac", "exp",
        "wood_chan"}.
    sigma : float, a scaling parameter.
    H : fractionnarity.
        ! If corr_type = "exp", H controls the fractionarity with a
        different range of values : H need to get into the thousands to
        have the same fractionarity as H ~ 0.3 with corr_type = "frac".
    v : drift, list of size dim.
    X_init : float or list, initial position of the random walk.
    dim : int, in {1,2,3}, the dimension of the random walk.
    """
    nb_point = int(T_max/dt)
    if corr_type == "frac":
        W = [fractional_correlated_noise(dt, H, nb_point-1) * sigma
             for _ in range(dim)]
    elif corr_type == "exp":
        W = [exponential_correlated_noise(dt, H, nb_point-1) * sigma
             for _ in range(dim)]
    elif corr_type == "wood_chan":
        W = [fractional_correlated_noise_wood_chan(dt, sigma, H, nb_point-1)
             for _ in range(dim)]
    else:
        raise ValueError(f'{corr_type} : unrecognized correlation type.')
    W = np.atleast_2d(W)[:, :nb_point-1]
    normed_pos_init = np.array(normalize_init(X_init, dim))
    drift = np.array(normalize_init(v, dim)) * dt * np.ones((nb_point-1, dim))
    W = W.cumsum(axis=1).T + normed_pos_init + drift
    W = np.concatenate((np.expand_dims(normed_pos_init, axis=0), W), axis=0)
    t = np.linspace(0, nb_point*dt, nb_point, endpoint=False)
    data = np.concatenate((np.expand_dims(t, axis=1), W), axis=1)
    return pd.DataFrame(data=data, columns=['t'] + SPACE_COLS[:dim])


# Confined Random Walk

def Ornstein_Uhlenbeck_integral(T_max=1, dt=1e-2, D=0.1, d_confinement=0.2,
                                nb_short=1, X_init=0, X_target=0, dim=2,
                                V_confinement=4):
    """
    Generate confined random walk following ornstein uhlenbeck process.
    follow equilibrium rule (D = kbT/gamma)
    center of the trap will be X_target to be modified if necessary
    avoid starting r_ini too far from x_init
    integral version.

    Parameters
    ----------
    T_max : float, the total time computed for which the random walk is
        computed.
    dt : float, time between each point output.
    D : float, the diffusivity of the random walk.
    d_confinement: typical scale of confinement corresponding to 4kbT.
    nb_short : int,  the number of points used between each point output, to
        avoid accumulating noise.
    X_init : float or list, initial position of the random walk.
    X_target : center of the confined walk.
    dim : int, in {1,2,3}, the dimension of the random walk.
    """
    # parameters definition
    gamma = 1 / D
    k_eff = 2 * V_confinement / (np.power(d_confinement, 2) * gamma)
    sigma = np.sqrt(2*D)
    nb_point = int(T_max/dt)
    nb_tot = (nb_point) * nb_short
    dt_short = dt/nb_short
    X_init = normalize_init(X_init, dim)

    t = np.arange(nb_tot)*dt_short
    X_target = np.array(normalize_init(X_target, dim))
    exp_k_eff = np.exp(-k_eff*t)
    exp_k_eff_2 = np.tile(exp_k_eff, (dim, 1)).T
    one_exp_k_eff_2 = np.ones((nb_tot, dim)) - exp_k_eff_2

    exp_2k_eff_minus_1 = np.exp(2 * k_eff * t) - 1.
    exp_2k_eff_minus_1_2 = np.tile(exp_2k_eff_minus_1, (dim, 1)).T
    exp_2k_eff_minus_1_2_diff = np.sqrt(np.diff(exp_2k_eff_minus_1_2, axis=0))

    term_gaussian = np.random.randn(nb_tot-1, dim)/np.sqrt(2*k_eff)
    joined_term = exp_2k_eff_minus_1_2_diff * term_gaussian
    joined_term_0 = np.insert(joined_term, 0, 0, axis=0)

    term1 = np.array(X_init)*exp_k_eff_2
    term2 = X_target*one_exp_k_eff_2
    term3 = sigma * exp_k_eff_2 * np.cumsum(joined_term_0, axis=0)
    X = term1 + term2 + term3
    data = np.stack((t,) + tuple(X[:, i] for i in range(dim))).T
    return pd.DataFrame(data=data, columns=['t'] + SPACE_COLS[:dim])


def Ornstein_Uhlenbeck_update(T_max=1, dt=1e-2, D=0.1, d_confinement=0.2,
                              nb_short=1, X_init=0, X_target=0, dim=2,
                              V_confinement=4):
    """
    Generate confined random walk following ornstein uhlenbeck process.
    follow equilibirum rule (D = kbT/gamma)
    center of the trap will be X_target to be modified if necessary
    avoid starting r_ini too far from x_init
    classical update version

    Parameters
    ----------
    T_max : float, the total time computed for which the random walk is
        computed.
    dt : float, time between each point output.
    D : float, the diffusivity of the random walk.
    d_confinement: typical scale of confinement corresponding to 4kbT
    nb_short : int,  the number of points used between each point output, to
        avoid accumulating noise.
    X_init : float or list, initial position of the random walk.
    X_target : center of the confined walk.
    dim : int, in {1,2,3}, the dimension of the random walk.
    """
    # parameters definition
    gamma = 1 / D
    k_eff = 2 * V_confinement / (np.power(d_confinement, 2) * gamma)
    sigma = np.sqrt(2 * D)
    nb_point = int(T_max/dt)
    nb_tot = (nb_point - 1) * nb_short
    dt_short = dt / nb_short
    k_eff_dt_short = k_eff * dt_short
    X_init = normalize_init(X_init, dim)
    t = np.arange(nb_tot)*dt_short
    X_target = np.array(normalize_init(X_target, dim))
    X_noise = np.sqrt(2*D*dt_short)*np.random.randn(nb_tot, dim)
    X = np.empty((nb_tot, dim))
    X[0] = X_init
    for i in range(1, nb_tot):
        X[i] = X[i-1] - k_eff_dt_short * (X[i-1] - X_target) + X_noise[i-1]
    data = np.stack((t,) + tuple(X[:, i] for i in range(dim))).T
    return pd.DataFrame(data=data, columns=['t'] + SPACE_COLS[:dim])


# Hidden Markov Model of diffusion, brownian movement.

def RW_HMM(T_max=1, dt=1e-2, D=np.array([0.001, 0.1]),
           T=np.array([[0.9, 0.1], [0.1, 0.9]]),
           p_init=np.array([0.5, 0.5]), v=None,
           nb_short=1, X_init=0, dim=2):
    """
    Generate diffusive random walk with diffusion following a markov model.

    Parameters
    ----------
    T_max : float, the total time computed for which the random walk is
        computed.
    dt : float, time between each point output.
    D : float, the k possible diffusivities of the random walk.
    T : transition matrix from one diffusive state to the other.
        Must have shape k x k and each row must sum to 1.
    p_init : initial state probability distribution, shape : k.
    v : speed, drift parameter. If not None, should be an array of size
        k x dim.
    nb_short : int,  the number of points used between each point output, to
        avoid accumulating noise.
    X_init : float or list, initial position of the random walk.
    dim : int, in {1,2,3}, the dimension of the random walk.
    """
    nb_point = int(T_max/dt)
    nb_tot = (nb_point - 1) * nb_short
    dt_short = dt/nb_short
    X = np.zeros((nb_point, dim)) + np.array(normalize_init(X_init, dim))
    states = np.zeros((nb_point, 1))
    Xi = X[0].copy()
    nstate = len(p_init)
    state = np.random.choice(nstate, size=1, p=p_init).item()
    states[0] = state
    if v is None:
        v = np.zeros((nstate, dim))
    for i in range(1, nb_tot+1):
        Xi += (np.sqrt(2*D[state]*dt_short)*np.random.randn(dim) +
               v[state]*dt_short)
        if i % nb_short == 0:
            X[i//nb_short] = Xi
            state = np.random.choice(nstate, size=1, p=T[state]).item()
            states[i//nb_short] = state
    t = np.linspace(0, nb_point*dt, nb_point, endpoint=False)
    data = np.concatenate((np.expand_dims(t, axis=1), X, states), axis=1)
    return pd.DataFrame(data=data, columns=(['t'] + SPACE_COLS[:dim] +
                                            ['state']))


# Confined movement.

def RW_circular_confinement(T_max=1, dt=1e-2, D=0.1, d_wall=0.2, nb_short=1,
                            X_init=0, dim=2, sigma_boundary=0.02,
                            V_confinement=10):
    # parameters definition
    # sigma_boundary : important parameter for regularization
    gamma_D = 1 / D
    nb_point = int(T_max / dt)
    nb_tot = (nb_point - 1) * nb_short
    dt_short = dt/nb_short
    X_init = normalize_init(X_init, dim)

    t = np.arange(nb_tot)*dt_short
    X_noise = np.sqrt(2*D*dt_short)*np.random.randn(nb_tot, dim)
    X = np.empty((nb_tot, dim))
    X[0] = X_init

    if dim == 1:
        for i in range(1, nb_tot):
            f_local = local_border_forces_1D(
                X[i-1], d_wall, sigma_boundary, V_confinement) * 1/gamma_D
            X[i] = X[i-1] + f_local*dt_short + X_noise[i-1]

    if dim == 2:
        for i in range(1, nb_tot):
            r, theta = cart2pol(X[i-1, 0], X[i-1, 1])
            f_local = local_border_forces_2D(
                r, d_wall, theta, sigma_boundary, V_confinement) * 1/gamma_D
            X[i] = X[i-1] + f_local*dt_short + X_noise[i-1]

    if dim == 3:
        for i in range(1, nb_tot):
            r, phi, theta = cart2spher(X[i-1, 0], X[i-1, 1],  X[i-1, 2])
            f_local = (local_border_forces_3D(
                r, d_wall, theta, phi, sigma_boundary, V_confinement) *
                1/gamma_D)
            X[i] = X[i-1] + f_local*dt_short + X_noise[i-1]

    X = X[::nb_short]
    t = t[::nb_short]
    data = np.stack((t,) + tuple(X[:, i] for i in range(dim))).T
    return pd.DataFrame(data=data, columns=['t'] + SPACE_COLS[:dim])


# Diffusing diffusivity

def RW_DD(T_max=1, dt=1e-2, nb_short=10, X_init=0, dim=2,
          mode_diffusion="chubynsky_slater", sigma_bruit=0.1, g_bruit=0.1,
          D_init=0.05, tau_bruit=5, n_dimension=2):
    """
    Generate diffusive random walk with diffusion itself following a random
    process.

    Parameters
    ----------
    T_max : float, the total time computed for which the random walk is
        computed.
    dt : float, time between each point output.
    nb_short : int,  the number of points used between each point output, to
        avoid accumulating noise.
    X_init : float or list, initial position of the random walk.
    dim : int, in {1,2,3}, the dimension of the random walk.
    mode_diffusion : string, the type of random process that diffusion follows
        Must be in {"chubynsky_slater", "langevin_square", "CIR"}.
    sigma_bruit : parameter equivalent to the diffusion in the random process
        of the diffusion.
    g_bruit : drift parameter in the random process of the diffusion.
    D_init : initial value taken by the diffusion of the random walk.
    tau_bruit : controls how much diffusion is above 0.
    n_dimension : dimension of the diffusion returned in case we want to
        generate a multi dimensional Diffusion process (with the langevin
        square distribution).
    """
    nb_point = int(T_max/dt)
    nb_tot = (nb_point-1) * nb_short
    dt_short = dt/nb_short
    Xi = np.random.randn(nb_tot, dim)
    X_init = normalize_init(X_init, dim)
    if mode_diffusion == "chubynsky_slater":
        D = give_diffusion_chubynsky_slater(nb_tot, dt_short, sigma_bruit,
                                            g_bruit, D_init)
    if mode_diffusion == "langevin_square":
        D = give_diffusion_langevin_square(n_dimension, nb_tot, dt_short,
                                           tau_bruit, sigma_bruit, D_init)
    if mode_diffusion == "CIR":
        D = give_diffusion_CIR(nb_tot, dt_short, tau_bruit, sigma_bruit,
                               D_init)
    D = npm.repmat(D, dim, 1).T
    X = np.cumsum(np.sqrt(2*D*dt_short)*Xi, axis=0) + np.array(X_init)
    X = X[::nb_short]
    X = np.insert(X, 0, X_init, axis=0)
    t = np.arange(nb_point)*dt
    data = np.stack((t,) + tuple(X[:, i] for i in range(dim))).T
    return pd.DataFrame(data=data, columns=['t'] + SPACE_COLS[:dim])


# SELF AVOIDING WALK
# code form: Computational investigations of folded self-avoiding
# walks related to protein folding
# Jacques Bahi, Christophe Guyeux, Kamel Mazouzi, Laurent Philippe
# Electronic Notes in Discrete Mathematics
# Volume 59, June 2017, Pages 37-50
# (slight modification)


# MISC functions with a global prm

def construction_sommets_graphe(pt_base, pt, d):
    global Graphe
    if d == 0:
        pt.reverse()
        pt = tuple(pt)
        if pt not in Dico_Des_Points:
            Graphe[pt] = []
    else:
        for dep in [-1, 0, 1]:
            construction_sommets_graphe(pt_base, pt + [pt_base[d - 1] + dep],
                                        d - 1)


def construction_aretes_graphe(dim):
    global Graphe
    for v1 in Graphe:
        for v2 in Graphe:
            if norme_manhattan(moinss(v1, v2, dim)) == 1:
                Graphe[v1].append(v2)
                Graphe[v2].append(v1)
    return Graphe


def calcul_nombres_infinis(dim, Deplacements):
    global Graphe
    pt_courant = Liste_Des_Points[-1]
    num_voisin = 0
    Graphe = {}
    resul = []
    while num_voisin < 2 * dim:
        candidat = pluss(pt_courant, Deplacements[num_voisin], dim)
        som = -1
        if candidat not in Dico_Des_Points:
            som = 0
            for ind in range(dim):
                petit_pt = enleve_coord(candidat, ind)
                if petit_pt not in Les_Minmax[ind]:
                    som += 2
                elif (candidat[ind] < Les_Minmax[ind][petit_pt][0] or
                      (candidat[ind] > Les_Minmax[ind][petit_pt][1])):
                    som += 1
            if som == 0 and candidat not in Dico_Ariane:
                if Graphe == {}:
                    construction_sommets_graphe(pt_courant, [], dim)
                    Graphe = construction_aretes_graphe(dim)
                    dest = intersection(Graphe, Dico_Ariane)
                chem = chemin(candidat, dest, Graphe)
                if chem is None:
                    som = -1
        resul.append(som)
        num_voisin += 1
    return resul


# Self Avoiding Random Walk useful functions


def nouveau_fil(pt_base):
    global Liste_Ariane
    global Dico_Ariane
    Liste_Ariane = [pt_base]
    Dico_Ariane = {pt_base: 0}


def racourcir_fil(index):
    while len(Liste_Ariane) > index:
        del Dico_Ariane[Liste_Ariane.pop()]


def ajout_dans_fil(pt):
    Dico_Ariane[tuple(pt)] = len(Liste_Ariane)
    Liste_Ariane.append(tuple(pt))


def insert_point(pt, dim):
    Liste_Des_Points.append(pt)
    Dico_Des_Points[pt] = None
    for index in range(dim):
        petit_pt = enleve_coord(pt, index)
        if petit_pt not in Les_Minmax[index]:
            Les_Minmax[index][petit_pt] = [pt[index], pt[index]]
        elif pt[index] < Les_Minmax[index][petit_pt][0]:
            Les_Minmax[index][petit_pt][0] = pt[index]
        elif pt[index] > Les_Minmax[index][petit_pt][1]:
            Les_Minmax[index][petit_pt][1] = pt[index]


def ajout_point(bias, dim, Deplacements):
    global Graphe
    pt_courant = Liste_Des_Points[-1]
    les_probas = calcul_proba(calcul_nombres_infinis(dim, Deplacements),
                              bias, dim)
    num_voisin = np.random.choice(les_probas)
    candidat = pluss(pt_courant, Deplacements[num_voisin], dim)
    for index in range(dim):
        petit_pt = enleve_coord(candidat, index)
        if ((petit_pt not in Les_Minmax[index]) or
                (candidat[index] > Les_Minmax[index][petit_pt][1])):
            insert_point(candidat, dim)
            nouveau_fil(voisin(candidat, index, 1))
            break
        elif candidat[index] < Les_Minmax[index][petit_pt][0]:
            insert_point(candidat, dim)
            nouveau_fil(voisin(candidat, index, -1))
            break
    else:
        if candidat in Dico_Ariane:
            insert_point(candidat, dim)
            racourcir_fil(Dico_Ariane[candidat])
        else:
            dest = intersection(Graphe, Dico_Ariane)
            chem = chemin(candidat, dest, Graphe)
            insert_point(candidat, dim)
            racourcir_fil(Dico_Ariane[chem[0]] + 1)
            for pt in chem[1: -1]:
                ajout_dans_fil(pt)


def RW_SAW(T_max=1, dt=1e-2, dr=1e-2, dim=2, bias=-1):
    """Generates a self avoiding random walk. This random walk nevers crosses a
    path where it has been.

    Parameters
    ----------
    T_max : float, the total time computed for which the random walk is
        computed.
    dt : float, time between each point output.
    dr : float, scale of the random walk.
    dim : int, in {1,2,3}, the dimension of the random walk.
    bias : scalar, somewhat corresponds to drift.
    """
    nb_point = int(T_max/dt)
    global Liste_Des_Points
    global Dico_Des_Points
    global Liste_Ariane
    global Dico_Ariane
    global Les_Minmax

    sigma = dr/4

    Origine = tuple([0] * dim)     # (0,..., 0)
    Liste_Des_Points = [Origine]
    Dico_Des_Points = {Origine: None}
    Petite_Origine = tuple([0] * (dim - 1))

    Les_Minmax = create_minmax(dim, Petite_Origine)

    Point_Un = list(Origine)
    Point_Un[0] = 1
    Point_Un = tuple(Point_Un)     # (1, 0,..., 0)

    Liste_Ariane = [Point_Un]
    Dico_Ariane = {Point_Un: 0}

    Deplacements = create_deplacement(dim)
    Graphe = {}

    Liste_Des_Points = [Origine]
    Dico_Des_Points = {Origine: None}
    Les_Minmax = []
    for index in range(dim):
        Les_Minmax.append({Petite_Origine: [0, 0]})
    Liste_Ariane = [Point_Un]
    Dico_Ariane = {Point_Un: 0}
    for nb in range(nb_point):
        ajout_point(bias, dim, Deplacements)

    noise_ = sigma*np.random.randn(nb_point, dim)
    X = np.asarray(Liste_Des_Points)[:-1] * dr + noise_
    t = np.arange(nb_point)*dt
    data = np.stack((t,) + tuple(X[:, i] for i in range(dim))).T
    return pd.DataFrame(data=data, columns=['t'] + SPACE_COLS[:dim])


def load_motif_into_memory(fractal_name):
    global MOTIF
    try:
        MOTIF = np.load(f'{FRACTAL_DIR}\{fractal_name}.npy')
    except:
        print('Could not load fractal, try to create fractal with pattern {}'
              .format(fractal_name))


def RW_on_fractal_pattern(T_max=1, dt=1e-2, dr=1e-2, X_init=0, v=None,
                          pattern=None, noise_frac=0.2, **kwargs):
    """Generates a random walk which moves on a fractal pattern. We add noise
    to make it look more realistic

    Parameters
    ----------
    T_max : float, the total time computed for which the random walk is
        computed.
    dt : float, time between each point output.
    dr : float, scale of the random walk.
    X_init : float or list, initial position of the random walk.
    v : drift, list of size dim.
    noise_frac : float > 0. The fraction of the movement which is due to noise.
        In practise, we add a gaussian noise with standard deviation equal to
        dr * noise_frac.

    Note : dim is fixed at 2 at the moment, but could be generalized to higher
        dimensions by changing patterns and the function update_position.
    """
    nb_point = int(T_max/dt)
    X = np.zeros((nb_point, 2), dtype=int)
    X[0] = int(np.floor(pattern.shape[0]/2))
    for i in range(1, nb_point):
        i_init = X[i-1, 0]
        j_init = X[i-1, 1]
        k, l, keep_going = update_position(pattern, i_init, j_init)
        if keep_going:
            X[i, 0] = k
            X[i, 1] = l
        else:
            break
    noise_ = noise_frac * dr * np.random.randn(nb_point, 2)
    X = (X - X[0, :] + normalize_init(X_init, 2)) * dr + noise_
    drift = np.array(normalize_init(v, 2)) * dt * np.ones((nb_point-1, 2))
    X[1:] += drift.cumsum(axis=0)
    t = np.arange(nb_point)*dt
    data = np.stack((t,) + tuple(X[:, i] for i in range(2))).T
    return pd.DataFrame(data=data, columns=['t'] + SPACE_COLS[:2])


# Diffusion-limited aggregation


def load_tree_into_memory(tree_name, DIR=TREE_DIR):
    global TREE
    try:
        TREE = np.load(f'{DIR}\\{tree_name}.npy')
    except:
        print('Could not load fractal, try to create fractal with pattern {}'
              .format(tree_name))


def RW_on_tree(T_max=1, dt=1e-2, dr=1e-2, X_init=0, v=None, tree=None,
               noise_frac=0.2, **kwargs):
    """Generates a random walk which moves on a tree. See Diffusion-limited
    aggregation.

    Parameters
    ----------
    T_max : float, the total time computed for which the random walk is
        computed.
    dt : float, time between each point output.
    dr : float, scale of the random walk.
    X_init : float or list, initial position of the random walk.
    v : drift, list of size dim.
    noise_frac : float > 0. The fraction of the movement which is due to noise.
        In practise, we add a gaussian noise with standard deviation equal to
        dr * noise_frac.

    Note : dim is fixed at 2 at the moment, but could be generalized to higher
        dimensions by changing patterns and the function update_position.
    """
    assert tree is not None, 'No tree given'
    nb_point = int(T_max/dt)
    X = np.zeros((nb_point, 2), dtype=int)
    X[0] = int(np.floor(tree.shape[0]/2))
    for i in range(1, nb_point):
        i_init = X[i-1, 0]
        j_init = X[i-1, 1]
        k, l, keep_going = update_position(tree, i_init, j_init)
        if keep_going:
            X[i, 0] = k
            X[i, 1] = l
        else:
            break
    noise_ = noise_frac * dr * np.random.randn(nb_point, 2)
    X = (X - X[0, :] + normalize_init(X_init, 2)) * dr + noise_
    drift = np.array(normalize_init(v, 2)) * dt * np.ones((nb_point-1, 2))
    X[1:] += drift.cumsum(axis=0)
    t = np.arange(nb_point)*dt
    data = np.stack((t,) + tuple(X[:, i] for i in range(2))).T
    return pd.DataFrame(data=data, columns=['t'] + SPACE_COLS[:2])



# -*- coding: utf-8 -*-

# Copyright © 2017-2021, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from .base import *
from .gradient import *
from warnings import warn
from math import pi, log
import numpy as np
import pandas as pd
from scipy.optimize import minimize
from collections import OrderedDict, namedtuple


setup = {'name': ('standard.d', 'smooth.d', 'regularized.d'),
    'provides': 'd',
    'arguments': OrderedDict((
        ('localization_error',  ('-e', dict(type=float, help='localization precision (see also sigma; default is 0.03)'))),
        ('diffusivity_prior',   ('-d', dict(type=float, help='prior on the diffusivity'))),
        ('jeffreys_prior',      ('-j', dict(action='store_true', help="Jeffreys' prior"))),
        ('min_diffusivity',     dict(type=float, help='minimum diffusivity value allowed')),
        ('max_iter',        dict(type=int, help='maximum number of iterations')),
        ('rgrad',       dict(help="alternative gradient for the regularization; can be 'delta'/'delta0' or 'delta1'")),
        ('tol',             dict(type=float, help='tolerance for scipy minimizer')))),
    'cell_sampling': 'group'}
setup_with_grad_arguments(setup)


Result = namedtuple('Result', 'result success message ncalls')

def interruptible_minimize(func, prms, args, **kwargs):
    callback_cache = dict(
            ncalls = 0,
            )

    def caching_callback(x, *args):
        callback_cache['result'] = x
        callback_cache['ncalls'] += 1
        return False

    kwargs['callback'] = caching_callback

    try:
        return minimize(func, prms, args=args, **kwargs)

    except KeyboardInterrupt:
        if 'result' in callback_cache:
            res = callback_cache['result']
            ctr = callback_cache['ncalls']
            ret = Result(res, False, 'interrupted', ctr)
            return ret
        else:
            raise


def smooth_d_neg_posterior(diffusivity, cells, sigma2, diffusivity_prior, \
    jeffreys_prior, dt_mean, min_diffusivity, index, reverse_index, grad_kwargs):
    """
    Adapted from InferenceMAP's *dDDPosterior* procedure:

    .. code-block:: c++

        for (int a = 0; a < NUMBER_OF_ZONES; a++) {
            ZONES[a].gradDx = dvGradDx(DD,a);
            ZONES[a].gradDy = dvGradDy(DD,a);
            ZONES[a].priorActive = true;
        }

        for (int z = 0; z < NUMBER_OF_ZONES; z++) {
            const double gradDx = ZONES[z].gradDx;
            const double gradDy = ZONES[z].gradDy;
            const double D = DD[z];

            for (int j = 0; j < ZONES[z].translocations; j++) {
                const double dt = ZONES[z].dt[j];
                const double dx = ZONES[z].dx[j];
                const double dy = ZONES[z].dy[j];
                const double Dnoise = LOCALIZATION_ERROR*LOCALIZATION_ERROR/dt;

                result += - log(4.0*PI*(D + Dnoise)*dt) - ( dx*dx + dy*dy)/(4.0*(D+Dnoise)*dt);
            }

            if (ZONES[z].priorActive == true) {
                result -= D_PRIOR*(gradDx*gradDx*ZONES[z].areaX + gradDy*gradDy*ZONES[z].areaY);
                if (JEFFREYS_PRIOR == 1) {
                    result += 2.0*log(D) - 2.0*log(D*ZONES[z].dtMean + LOCALIZATION_ERROR*LOCALIZATION_ERROR);
                }
            }
        }

        return -result;

    """
    if min_diffusivity is not None:
        observed_min = np.min(diffusivity)
        if observed_min < min_diffusivity and not np.isclose(observed_min, min_diffusivity):
            warn(DiffusivityWarning(observed_min, min_diffusivity))
    noise_dt = sigma2
    result = 0.
    for j, i in enumerate(index):
        cell = cells[i]
        n = len(cell)
        # posterior calculations
        if cell.cache is None:
            cell.cache = dict(dr2=None)
        if cell.cache['dr2'] is None:
            cell.cache['dr2'] = np.sum(cell.dr * cell.dr, axis=1) # dx**2 + dy**2 + ..
        D_dt = 4. * (diffusivity[j] * cell.dt + noise_dt) # 4*(D+Dnoise)*dt
        result += n * log(pi) + np.sum(np.log(D_dt)) # sum(log(4*pi*Dtot*dt))
        result += np.sum(cell.cache['dr2'] / D_dt) # sum((dx**2+dy**2+..)/(4*Dtot*dt))
        # prior
        if diffusivity_prior:
            # gradient of diffusivity
            gradD = cells.grad(i, diffusivity, reverse_index, **grad_kwargs)
            if gradD is not None:
                result += diffusivity_prior * cells.grad_sum(i, gradD * gradD)
    if jeffreys_prior:
        result += 2. * np.sum(np.log(diffusivity * dt_mean + sigma2))
    return result


def d_neg_posterior1(diffusivity, cells, sigma2, diffusivity_prior, \
    jeffreys_prior, dt_mean, min_diffusivity, index, reverse_index, grad_kwargs):
    """
    Similar to :func:`smooth_d_neg_posterior`.
    The smoothing prior features an alternative spatial "gradient" implemented using
    :meth:`~tramway.inference.base.Distributed.local_variation` instead of
    :meth:`~tramway.inference.base.Distributed.grad`.
    """
    if min_diffusivity is not None:
        observed_min = np.min(diffusivity)
        if observed_min < min_diffusivity:# and not np.isclose(observed_min, min_diffusivity):
            warn(DiffusivityWarning(observed_min, min_diffusivity))
    noise_dt = sigma2
    result = 0.
    for j, i in enumerate(index):
        cell = cells[i]
        n = len(cell)
        # posterior calculations
        if cell.cache is None:
            cell.cache = dict(dr2=None)
        if cell.cache['dr2'] is None:
            cell.cache['dr2'] = np.sum(cell.dr * cell.dr, axis=1) # dx**2 + dy**2 + ..
        D_dt = 4. * (diffusivity[j] * cell.dt + noise_dt) # 4*(D+Dnoise)*dt
        result += n * log(pi) + np.sum(np.log(D_dt)) # sum(log(4*pi*Dtot*dt))
        result += np.sum(cell.cache['dr2'] / D_dt) # sum((dx**2+dy**2+..)/(4*Dtot*dt))
        # prior
        if diffusivity_prior:
            # gradient of diffusivity
            deltaD = cells.local_variation(i, diffusivity, reverse_index, **grad_kwargs)
            if deltaD is not None:
                result += diffusivity_prior * cells.grad_sum(i, deltaD * deltaD)
    if jeffreys_prior:
        result += 2. * np.sum(np.log(diffusivity * dt_mean + sigma2))
    return result


def infer_smooth_D(cells, diffusivity_prior=None, jeffreys_prior=None, \
    min_diffusivity=None, max_iter=None, epsilon=None, rgrad=None, verbose=False, **kwargs):

    # initial values
    localization_error = cells.get_localization_error(kwargs, 0.03, True)
    index, reverse_index, n, dt_mean, D_initial, min_diffusivity, D_bounds, _ = \
        smooth_infer_init(cells, min_diffusivity=min_diffusivity, jeffreys_prior=jeffreys_prior,
        sigma2=localization_error)

    # gradient options
    grad_kwargs = get_grad_kwargs(kwargs, epsilon=epsilon)

    # parametrize the optimization algorithm
    default_lBFGSb_options = dict(maxiter=1e3, maxfun=1e10, ftol=1e-6)
    # in L-BFGS-B the number of iterations is usually very low (~10-100) while the number of
    # function evaluations is much higher (~1e4-1e5);
    # with maxfun defined, an iteration can stop anytime and the optimization may terminate
    # with an error message
    if min_diffusivity is None:
        options = {}
    else:
        kwargs['bounds'] = D_bounds
        options = dict(default_lBFGSb_options)
    options.update(kwargs.pop('options', {}))
    if max_iter:
        options['maxiter'] = max_iter
    if verbose:
        options['disp'] = verbose
    if options:
        kwargs['options'] = options

    # posterior function
    if rgrad in ('delta','delta0','delta1'):
        fun = d_neg_posterior1
    else:
        if rgrad not in (None, 'grad', 'grad1', 'gradn'):
            warn('unsupported rgrad: {}'.format(rgrad), RuntimeWarning)
        fun = smooth_d_neg_posterior

    args = (cells, localization_error, diffusivity_prior, jeffreys_prior, dt_mean, min_diffusivity, index, reverse_index, grad_kwargs)

    # run the optimization
    result = interruptible_minimize(fun, D_initial, args=args, **kwargs)
    if not (result.success or verbose):
        warn('{}'.format(result.message), OptimizationWarning)

    # format the result
    D = result.x
    D = pd.DataFrame(D, index=index, columns=['diffusivity'])

    return D




# -*- coding: utf-8 -*-

# Copyright © 2020-2021, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from ..attribute import *
from ..artefact import *
from ..roi import HasROI
from .abc import *
from collections.abc import Sequence, Set
import os.path
from tramway.core.xyt import load_xyt, load_mat, discard_static_trajectories
from tramway.core.analyses import base, lazy
from tramway.core.analyses.auto import Analyses, AutosaveCapable
from tramway.core.hdf5.store import load_rwa
from rwa.lazy import islazy
from tramway.core.exceptions import RWAFileException
from math import sqrt
import numpy as np
import pandas as pd
import copy
import glob as _glob


def compute_dtypes(df, precision):
    """
    Returns a dictionnary of dtypes to align the numerical precision of
    a :class:`pandas.DataFrame` data.
    """
    if isinstance(precision, dict):
        dtypes = precision
    else:
        if precision == 'half':
            nbytes = '2'
        elif precision == 'single':
            nbytes = '4'
        elif precision == 'double':
            nbytes = '8'
        else:
            raise ValueError('precision not supported: {}'.format(precision))
        dtypes = {}
        for col, dtype in df.dtypes.items():
            dtype = np.dtype(dtype).str
            if dtype[0] == '<' and dtype[1] in 'uif' and dtype[2] != nbytes:
                dtypes[col] = np.dtype(dtype[:2]+nbytes)
    return dtypes


def glob(pattern):
    """ Same as :func:`glob.glob`, with support for tilde.

    :func:`os.path.expanduser` is applied to `pattern`,
    and then the expanded part is replaced back with
    tilde in the resulting paths."""
    if pattern.startswith('~'):
        home = os.path.expanduser('~')
        files = []
        for f in _glob.glob(os.path.expanduser(pattern)):
            if f.startswith(home):
                f = '~' + f[len(home):]
            files.append(f)
    else:
        files = _glob.glob(pattern)
    return files


class SPTParameters(object):
    """
    Mixin class for storing common experimental parameters, including the
    localization error, time interval and temperature.

    Children classes should define the :attr:`_frame_interval`,
    :attr:`_localization_error` and :attr:`_temperature` attributes,
    or overload the :attr:`frame_interval`, :attr:`localization_error`
    and :attr:`temperature` properties.

    Default values should be :const:`None`.
    """
    __slots__ = ()
    def __init__(self, localization_precision=None, localization_error=None):
        if localization_precision is None:
            if localization_error is not None:
                self.localization_error = localization_error
        elif localization_error is None:
            self.localization_precision = localization_precision
        else:
            raise ValueError('both error and precision are defined; choose either one')
    @classmethod
    def __parse__(cls, kwargs):
        return (
            kwargs.pop('localization_precision', None),
            kwargs.pop('localization_error', None),
            )
    @property
    def localization_error(self):
        return self._localization_error
    @localization_error.setter
    def localization_error(self, err):
        if err is None:
            err = 0.
        self._localization_error = err
    localization_error.__doc__ = SPTData.localization_error.__doc__
    @property
    def localization_precision(self):
        r"""
        *float*: Localization precision in :math:`\mu m`;
            :attr:`localization_error` :math:`\sigma^2` is affected by
            :attr:`localization_precision` :math:`\sigma` and vice versa
        """
        if self.localization_error is None:
            return None
        else:
            return sqrt(self.localization_error)
    @localization_precision.setter
    def localization_precision(self, pr):
        err = pr if pr is None else pr*pr
        self.localization_error = err
    def discard_static_trajectories(self, dataframe, min_msd=None, **kwargs):
        if min_msd is None:
            min_msd = self.localization_error
        return discard_static_trajectories(dataframe, min_msd, **kwargs)
    discard_static_trajectories.__doc__ = SPTData.discard_static_trajectories.__doc__
    @property
    def frame_interval(self):
        if self._frame_interval is None:
            t = self.dataframe['t']
            self._frame_interval = t.diff().median()
        return self._frame_interval
    @frame_interval.setter
    def frame_interval(self, dt):
        self._frame_interval = dt
    frame_interval.__doc__ = SPTData.frame_interval.__doc__
    @property
    def time_step(self):
        """
        *float*: Alias for the :attr:`frame_interval` property
        """
        return self.frame_interval
    @time_step.setter
    def time_step(self, dt):
        self.frame_interval = dt
    @property
    def dt(self):
        return self.frame_interval
    @dt.setter
    def dt(self, dt):
        self.frame_interval = dt
    dt.__doc__ = time_step.__doc__
    @property
    def temperature(self):
        """ *float*: Temperature in *K*; can be set as a *value*-*unit* pair,
            with *unit* any of :const:`'K'`, :const:`'C'` and :const:`'F'`
        """
        return self._temperature
    @temperature.setter
    def temperature(self, T):
        if isinstance(T, (tuple, list)):
            T, unit = T
            if T is not None:
                if unit == 'C':
                    T = 273 + T
                elif unit == 'F':
                    T = 273 + (T - 32) / 1.8
                elif unit != 'K':
                    raise ValueError(f"temperature unit not supported: '{unit}'")
        self._temperature = T
    def set_precision(self, precision):
        """
        Sets the numerical precision of the raw data.

        Arguments:

            precision (*dict* or *str*): any of :const:`'half'`, :const:`'single'`, :const:`'double'`,
                or a dictionnary of dtypes with column names as keys,
                as admitted by :meth:`pandas.DataFrame.astype`.

        """
        dtypes = compute_dtypes(self.dataframe, precision)
        self.dataframe = self.dataframe.astype(dtypes)


def _normalize(p):
    return os.path.expanduser(os.path.normpath(p))


class SPTDataIterator(AnalyzerNode, SPTParameters):
    """ Partial implementation for multi-item :class:`SPTData`.

    Children classes must implement the :meth:`__iter__` method.
    """
    __slots__ = ('_bounds',)
    def __init__(self, **kwargs):
        prms = SPTParameters.__parse__(kwargs)
        AnalyzerNode.__init__(self, **kwargs)
        self._bounds = None
        SPTParameters.__init__(self, *prms)
    @property
    def localization_error(self):
        it = iter(self)
        sigma2 = next(it).localization_error
        while True:
            try:
                _s2 = next(it).localization_error
            except StopIteration:
                break
            else:
                if _s2 != sigma2:
                    raise AttributeError('not all the data blocks share the same localization error')
        return sigma2
    @localization_error.setter
    def localization_error(self, sigma2):
        for f in self:
            f.localization_error = sigma2
    @property
    def frame_interval(self):
        it = iter(self)
        dt = next(it).frame_interval
        while True:
            try:
                _dt = next(it).frame_interval
            except StopIteration:
                break
            else:
                _delta = dt - _dt
                if 1e-12 < _delta*_delta:
                    raise AttributeError('not all the data blocks share the same frame interval (dt)')
        return dt
    @frame_interval.setter
    def frame_interval(self, dt):
        for f in self:
            f.frame_interval = dt
    @property
    def dt(self):
        return self.frame_interval
    @dt.setter
    def dt(self, dt):
        self.frame_interval = dt
    @property
    def time_step(self):
        return self.frame_interval
    @time_step.setter
    def time_step(self, dt):
        self.frame_interval = dt
    @property
    def temperature(self):
        it = iter(self)
        T = next(it).temperature
        while True:
            try:
                _T = next(it).temperature
            except StopIteration:
                break
            else:
                if T != _T:
                    raise AttributeError('not all the data blocks share the same temperature')
        return T
    @temperature.setter
    def temperature(self, T):
        for f in self:
            f.temperature = T
    def as_dataframes(self, source=None, return_index=False):
        """ Generator function; yields :class:`pandas.DataFrame` objects.
        
        `source` can be a source name (filepath) or a boolean function
        that takes a source string as input argument.
        """
        for f in self.filter_by_source(source, return_index):
            yield f.dataframe
    def filter_by_source(self, source_filter, return_index=False):
        """ Generator function; similar to :meth:`__iter__`;
        yields :class:`SPTDataItem` objects.

        *source* can be a single `str` value, or a set of `str` values,
        or a sequence of `str` values (the order is followed),
        or a `callable` that takes a `str` value and returns a `bool` value.
        """
        if return_index:
            _out = lambda i, f: i, f
        else:
            _out = lambda _, f: f
        if source_filter is None:
            if return_index:
                yield from enumerate(self)
            else:
                yield from self
        elif callable(source_filter):
            for i, f in enumerate(self):
                if source_filter(f.source):
                    yield _out(i, f)
        else:
            if isinstance(source_filter, str):
                sources = [_normalize(source_filter)]
            elif isinstance(source_filter, Sequence):
                visited = dict()
                yielded = set()
                for _p in source_filter:
                    p = _normalize(_p)
                    i, it = -1, iter(self)
                    try:
                        while True:
                            i += 1
                            f = next(it)
                            try:
                                s = visited[i]
                            except KeyError:
                                visited[i] = s = _normalize(f.source)
                            if s == p:
                                if i in yielded:
                                    raise ValueError('duplicate source: {}'.format(_p))
                                yield _out(i, f)
                                yielded.add(i)
                                break
                    except StopIteration:
                        raise ValueError('cannot find source: {}'.format(_p))
                return
            elif isinstance(source_filter, Set):
                sources = set([ _normalize(p) for p in source_filter ])
            for i, f in enumerate(self):
                p = _normalize(f.source)
                if p in sources:
                    yield _out(i, f)
    def discard_static_trajectories(self, dataframe=None, min_msd=None, **kwargs):
        if dataframe is None:
            for data in self:
                data.discard_static_trajectories(min_msd=min_msd, **kwargs)
        else:
            return SPTParameters.discard_static_trajectories(self, dataframe, min_msd, **kwargs)
    @property
    def bounds(self):
        if self._bounds is None:
            for f in self:
                _bounds = f.bounds
                if self._bounds is None:
                    self._bounds = _bounds
                else:
                    self._bounds.loc['min'] = np.minimum(self._bounds.loc['min'], _bounds.loc['min'])
                    self._bounds.loc['max'] = np.minimum(self._bounds.loc['max'], _bounds.loc['max'])
        return self._bounds
    bounds.__doc__ = SPTData.bounds.__doc__
    def self_update(self, new_self):
        """
        """
        if callable(new_self):
            f = new_self
            new_self = f(self)
        if isinstance(self, Proxy) and not isinstance(new_self, Proxy):
            new_self = type(self)(new_self)
        if new_self is not self:
            self._parent._spt_data = new_self
            new_self._parent = self._parent
            try:
                roi_central = next(iter(self)).roi._global
                roi_central.reset()
            except AttributeError:
                pass
            else:
                for f in new_self:
                    roi_central._register_decentralized_roi(f)
        return new_self
    def reload_from_rwa_files(self, skip_missing=False):
        """
        Reloads the SPT data and analysis tree from the corresponding rwa files.

        The rwa file that corresponds to an SPT file should be available at the
        same path with the *.rwa* extension instead of the SPT file's extension.

        This method is known to fail with a :class:`TypeError` exception in cases
        where not any matching *.rwa* file can be found.

        .. note::

            As this operation modifies the SPT data `source` and `filepath` attributes,
            aliases should be favored when identifying or filtering SPT data items.

        """
        return RWAFiles.reload_from_rwa_files(self, skip_missing=skip_missing)
    def set_precision(self, precision):
        for f in self:
            f.set_precision(precision)
    def close(self):
        for f in self:
            f.close()


class SPTDataInitializer(Initializer):
    """
    Initial value for the :class:`~tramway.analyzer.RWAnalyzer`
    :attr:`~tramway.analyzer.RWAnalyzer.spt_data` attribute.

    *from_...* methods alters the parent attribute which specializes
    into an initialized :class:`SPTData` object.
    """
    __slots__ = ()
    def from_ascii_file(self, filepath):
        """
        Sets a text file as the source of SPT data.

        Note that data loading is NOT performed while calling this method.
        Loading is postponed until the data is actually required.
        This lets additional arguments to be provided to the
        :class:`~tramway.analyzer.RWAnalyzer` :attr:`~tramway.analyzer.RWAnalyzer.spt_data`
        attribute before the data are loaded.

        See also :class:`StandaloneSPTAsciiFile`.
        """
        self.specialize( StandaloneSPTAsciiFile, filepath )
    def from_ascii_files(self, filepattern):
        """
        Sets text files, which paths match with a pattern, as the source of SPT data.

        `filepattern` is a standard filepath with the :const:`'*'` placeholder.
        For example:  `'dataset/*.txt'`

        The parts of the filename that match the placeholder are used as keys.

        Note that data loading is NOT performed while calling this method.
        Loading is postponed until the data is actually required.
        This lets additional arguments to be provided to the
        :class:`~tramway.analyzer.RWAnalyzer` :attr:`~tramway.analyzer.RWAnalyzer.spt_data`
        attribute before the data are loaded.

        See also :class:`SPTAsciiFiles`.
        """
        self.specialize( SPTAsciiFiles, filepattern )
    def from_dataframe(self, df):
        """
        See also :class:`StandaloneSPTDataFrame`.
        """
        self.specialize( StandaloneSPTDataFrame, df )
    def from_dataframes(self, dfs):
        """
        See also :class:`SPTDataFrames`.
        """
        self.specialize( SPTDataFrames, dfs )
    def from_mat_file(self, filepath):
        """
        Sets a MatLab V7 file as the source of SPT data.

        Similarly to :meth:`from_ascii_file`, data loading is lazy.

        See also :class:`StandaloneSPTMatFile`.
        """
        self.specialize( StandaloneSPTMatFile, filepath )
    def from_mat_files(self, filepattern):
        """
        Sets MatLab V7 files, which paths match with a pattern, as the source of SPT data.

        `filepattern` is a standard filepath with the :const:`'*'` placeholder.
        For example:  `'datasets/*.txt'`

        The parts of the filename that match the placeholder are used as keys.

        Similarly to :meth:`from_ascii_files`, data loading is lazy.

        See also :class:`SPTMatFiles`.
        """
        self.specialize( SPTMatFiles, filepattern )
    def from_rwa_file(self, filepath):
        """
        Similar to :meth:`from_ascii_file`, for *.rwa* files.

        See also :class:`StandaloneRWAFile`.
        """
        self.specialize( StandaloneRWAFile, filepath )
    def from_rwa_files(self, filepattern):
        """
        Similar to :meth:`from_ascii_files`, for *.rwa* files.

        See also :class:`RWAFiles`.
        """
        self.specialize( RWAFiles, filepattern )
    def from_rw_generator(self, generator):
        """
        A random walk generator features a :meth:`generate` method.

        See also :class:`RWGenerator`.
        """
        self.specialize( RWGenerator, generator )
    def from_analysis_tree(self, analyses, copy=False):
        self.specialize( RWAnalyses, analyses, copy )
    def from_tracker(self):
        """ This initializer method does not need to be called;
        The :class:`~tramway.analyzer.RWAnalyzer`
        :attr:`~tramway.analyzer.RWAnalyzer.tracker` attribute does this automatically.

        See also :class:`TrackerOutput`.
        """
        self.specialize( TrackerOutput )



class StandaloneDataItem(object):
    """
    Partial implementation for single data item
    :class:`~tramway.analyzer.spt_data.SPTData` attribute.
    """
    __slots__ = ()
    def __len__(self):
        return 1
    def __iter__(self):
        yield self
    def as_dataframes(self, source=None):
        return SPTDataIterator.as_dataframes(self, source)
    def self_update(self, new_self):
        if callable(new_self):
            f = new_self
            new_self = f(self)
        if isinstance(self, Proxy) and not isinstance(new_self, Proxy):
            new_self = type(self)(new_self)
        if new_self is not self:
            self._parent._spt_data = new_self
            new_self._parent = self._parent
            try:
                roi_central = next(iter(self)).roi._global
                roi_central.reset()
            except AttributeError:
                pass
            else:
                for f in new_self:
                    roi_central._register_decentralized_roi(f)
        return new_self
    def reload_from_rwa_files(self, skip_missing=False):
        return StandaloneRWAFile.reload_from_rwa_files(self, skip_missing=skip_missing)


class HasAnalysisTree(HasROI):
    """
    Partial implementation for :class:`SPTData` that complements :class:`SPTParameters`.
    """
    __slots__ = ('_analyses', '_frame_interval_cache',
            '_localization_error_cache', '_temperature_cache')
    def __init__(self, df=None, **kwargs):
        self._frame_interval_cache = None
        self._localization_error_cache = None
        self._temperature_cache = None
        HasROI.__init__(self, **kwargs)
        self._analyses = None
        self.analyses = Analyses(df, standard_metadata(), autosave=True)
        self._analyses.hooks.append(lambda _: self.commit_cache(autoload=True))
    @property
    def _frame_interval(self):
        return self._frame_interval_cache
    @_frame_interval.setter
    def _frame_interval(self, dt):
        self._frame_interval_cache = dt
        if not islazy(self._dataframe) and self._dataframe is not None:
            self._dataframe.frame_interval = dt
    @property
    def _localization_error(self):
        return self._localization_error_cache
    @_localization_error.setter
    def _localization_error(self, err):
        self._localization_error_cache = err
        if not islazy(self._dataframe) and self._dataframe is not None:
            self._dataframe.localization_error = err
    @property
    def _temperature(self):
        return self._temperature_cache
    @_temperature.setter
    def _temperature(self, T):
        self._temperature_cache = T
        if not islazy(self._dataframe) and self._dataframe is not None:
            self._dataframe.temperature = T
    def commit_cache(self, autoload=False):
        """
        Pushes the cached parameters into the :attr:`dataframe` object.
        """
        if islazy(self._dataframe) or self._dataframe is None:
            if not autoload:
                raise RuntimeError('the SPT data has not been loaded')
        self.dataframe.frame_interval = self._frame_interval_cache
        self.dataframe.localization_error = self._localization_error_cache
        self.dataframe.temperature = self._temperature_cache
    def clear_cache(self):
        """
        Clears the cached values and reads from the :attr:`dataframe` object.
        """
        self._frame_interval_cache = None
        self._localization_error_cache = None
        self._temperature_cache = None
        if not islazy(self._dataframe) and self._dataframe is not None:
            try:
                self._frame_interval_cache = self._dataframe.frame_interval
            except AttributeError:
                pass
            try:
                self._localization_error_cache = self._dataframe.localization_error
            except AttributeError:
                pass
            try:
                self._temperature_cache = self._dataframe.temperature
            except AttributeError:
                pass
    def check_cache(self, _raise=AttributeError):
        """
        Checks the parameter cache integrity.

        If differences are found with the values in :attr:`dataframe`,
        :meth:`check_cache` raises an exception of type `_raise`.

        If `_raise` is :const:`None` or :const:`False`, then :meth:`check_cache` returns
        a `bool` instead, that is :const:`False` if the cache is alright,
        :const:`True` otherwise.
        If `_raise` is :const:`True`, then :meth:`check_cache` returns :const:`True` if
        the cache is alright, :const:`False` otherwise.
        """
        if _raise is None:
            _raise = False
        if isinstance(_raise, bool):
            _return = _raise
            _raise = None
        elif isinstance(_raise, Exception):
            _return = None
        elif isinstance(_raise, type) and issubclass(_raise, Exception):
            _return = None
            _raise = _raise('cache integrity is compromised')
        else:
            raise TypeError('unsupported type for second argument: {}'.format(type(_raise)))
        if not islazy(self._dataframe) and self._dataframe is not None:
            try:
                ok = self._frame_interval_cache == self._dataframe.frame_interval
            except AttributeError:
                pass
            else:
                if not ok:
                    if _return is None: # _raise is not None
                        raise _raise from None
                    else:
                        return not _return
            try:
                ok = self._localization_error_cache == self._dataframe.localization_error
            except AttributeError:
                pass
            else:
                if not ok:
                    if _return is None: # _raise is not None
                        raise _raise from None
                    else:
                        return not _return
            try:
                ok = self._temperature_cache == self._dataframe.temperature
            except AttributeError:
                pass
            else:
                if not ok:
                    if _return is None: # _raise is not None
                        raise _raise from None
                    else:
                        return not _return
        # the cache is alright
        if _return is not None:
            return _return
    def set_analyses(self, tree):
        if isinstance(tree, AutosaveCapable):
            autosaver = tree
            tree = autosaver.analyses.statefree()
        else:
            autosaver = Analyses(tree, autosave=True)
        if self._analyses is None:
            self._analyses = autosaver
        else:
            # in the very special case `self` results from a call to `_RWAFile.__reload__`:
            try:
                rwa_file = self._analyses.rwa_file
            except AttributeError:
                rwa_file = None
                autosave = None
            else:
                autosave = self._analyses.autosave
            self._analyses = autosaver
            if rwa_file is not None:
                self._analyses.rwa_file = rwa_file
            if autosave is not None:
                self._analyses.autosave = autosave
        #
        err_cache = self._localization_error_cache
        dt_cache = self._frame_interval_cache
        T_cache = self._temperature_cache
        if islazy(tree._data):
            store = tree._data.store
            record = store.getRecord(tree._data.locator, store.store)
            try:
                err = store.peek('localization_error', record)
            except KeyError:
                err = None
            try:
                dt = store.peek('frame_interval', record)
            except KeyError:
                dt = None
            try:
                T = store.peek('temperature', record)
            except KeyError:
                T = None
        else:
            df = tree.data
            try:
                T = df.temperature
            except AttributeError:
                T = None
            try:
                dt = df.frame_interval
            except AttributeError:
                dt = None
            try:
                err = df.localization_error
            except AttributeError:
                err = None
        if err is not None:
            if not (err_cache is None or err_cache == err):
                self.logger.warning("localization error does not match with record: {} != {}".format(err_cache, err))
            self._localization_error_cache = err
        if dt is not None:
            if not (dt_cache is None or dt_cache == dt):
                self.logger.warning("frame interval does not match with record: {} != {}".format(dt_cache, dt))
            self._frame_interval_cache = dt
        if T is not None:
            if not (T_cache is None or T_cache == T):
                self.logger.warning(f"temperature does not match with record: {T_cache} != {T}")
            self._temperature_cache = T
    @property
    def analyses(self):
        return self._analyses
    @analyses.setter
    def analyses(self, tree):
        self.set_analyses(tree)
    @property
    def _dataframe(self):
        return self._analyses._data
    @_dataframe.setter
    def _dataframe(self, df):
        self._analyses._data = df
    @property
    def dataframe(self):
        return self.analyses.data
    def close(self):
        if self._analyses is not None:
            tree = self
            while True:
                try:
                    wrapped_tree = tree._analyses
                except AttributeError:
                    break
                else:
                    tree = wrapped_tree
            try:
                wrapped_tree = tree.statefree()
            except AttributeError:
                pass
            else:
                tree = wrapped_tree
            tree.terminate()


class _SPTDataFrame(HasAnalysisTree, SPTParameters):
    """
    Basis for all concrete :class:`SPTDataItem` classes.
    """
    __slots__ = ('_source',)
    def __init__(self, df, source=None, **kwargs):
        prms = SPTParameters.__parse__(kwargs)
        self._source = source
        HasAnalysisTree.__init__(self, df, **kwargs)
        SPTParameters.__init__(self, *prms)
    @property
    def reified(self):
        return self._dataframe is not None
    @property
    def columns(self):
        return self.dataframe.columns
    def reset_origin(self, columns=None):
        if columns is None:
            columns = [ col for col in ['x', 'y', 'z', 't'] if col in self.columns ]
        origin = self.dataframe[columns].min().values
        self.dataframe[columns] -= origin
    def discard_static_trajectories(self, dataframe=None, min_msd=None, **kwargs):
        inplace = dataframe is None
        if inplace:
            dataframe = self._dataframe
        dataframe = SPTParameters.discard_static_trajectories(self, dataframe, min_msd, **kwargs)
        if inplace:
            self._dataframe = dataframe
        else:
            return dataframe
    @property
    def source(self):
        return self._source
    @property
    def bounds(self):
        _min = self.dataframe.min(axis=0)
        _max = self.dataframe.max(axis=0)
        _bounds = np.stack((_min, _max), axis=0)
        _bounds = pd.DataFrame( _bounds, columns=_min.index, index=['min','max'])
        return _bounds
    def set_precision(self, precision):
        dtypes = compute_dtypes(self.dataframe, precision)
        self._dataframe = self.dataframe.astype(dtypes)
    def to_ascii_file(self, filepath, columns=None, header=True, float_format='%.4f', **kwargs):
        """
        Exports the data to text file.

        Beware: whitespaces in the column names should be avoided with
        ``header=True``

        Arguments:

            filepath (str or Path): output filepath.

            columns (sequence of *str*): columns to be exported.

            header (bool): print column names on the first line.

            float_format (str): see also :meth:`pandas.DataFrame.to_csv`.

        Additional keyword arguments are passed to :meth:`pandas.DataFrame.to_csv`.
        """
        df = self.dataframe
        if columns:
            df = df[columns]
        for arg in ('sep', 'index'):
            try:
                kwargs.pop(arg)
            except KeyError:
                pass
            else:
                self.logger.warning("ignoring argument '{}'".format(arg))
        if header and any([ ' ' in col for col in df.columns ]):
            self.logger.warning('whitespaces in the column names')
        df.to_csv(str(filepath), sep='\t', index=False, header=header,
                float_format=float_format, **kwargs)
    def to_rwa_file(self, filepath, **kwargs):
        """
        Exports the analysis tree to file.

        Calls :func:`~tramway.core.hdf5.store.save_rwa` with argument
        ``compress=False`` unless explicitly set.
        """
        if self.analyses.data is None:
            raise ValueError('no data available')
        try:
            analyses = self.analyses.analyses.statefree()
        except AttributeError:
            analyses = self.analyses
        if 'compress' not in kwargs:
            kwargs['compress'] = False
        save_rwa(filepath, analyses, **kwargs)
    def set_analyses(self, tree):
        assert bool(tree.metadata)
        if self.source is not None:
            tree.metadata['datafile'] = self.source
        HasAnalysisTree.set_analyses(self, tree)
    def add_sampling(self, sampling, label, comment=None):
        analyses = self.analyses
        label = analyses.autoindex(label)
        subtree = lazy.Analyses(sampling, standard_metadata())
        analyses.add(subtree, label, comment)
        return label
    def get_sampling(self, label):
        return Analysis.get_analysis(self.analyses, label)
    def autosaving(self, *args, overwrite=True, **kwargs):
        assert isinstance(self.analyses, AutosaveCapable)
        if not self.analyses.rwa_file:
            if self.source:
                self.analyses.rwa_file = os.path.splitext(self.source)[0]+'.rwa'
                if self.analyses.rwa_file == self.source:
                    if not overwrite:
                        i = 0
                        while os.path.isfile(os.path.expanduser(self.analyses.rwa_file)):
                            i += 1
                            self.analyses.rwa_file = '{}-{:d}.rwa'.format(os.path.splitext(self.source)[0],i)
            else:
                self.logger.warning('no output filename defined')
        return self.analyses.autosaving(*args, **kwargs)
    @property
    def _mpl_impl(self):
        from .mpl import Mpl
        return Mpl
    @property
    def mpl(self):
        """ tramway.analyzer.spt_data.mpl.Mpl: Matplotlib utilities """
        return self._mpl_impl(self)


class SPTDataFrame(_SPTDataFrame):
    __slots__ = ()

SPTDataItem.register(SPTDataFrame)


class StandaloneSPTDataFrame(_SPTDataFrame, StandaloneDataItem):
    """
    :class:`SPTData` attribute for single dataframes.
    """
    __slots__ = ()

SPTData.register(StandaloneSPTDataFrame)


class SPTDataFrames(SPTDataIterator):
    """
    :class:`SPTData` attribute for multiple dataframes.
    """
    __slots__ = ('_dataframes',)
    def __init__(self, dfs, **kwargs):
        SPTDataIterator.__init__(self, **kwargs)
        self.dataframes = dfs
    @property
    def dataframes(self):
        return [ df.dataframe for df in self._dataframes ]
    @dataframes.setter
    def dataframes(self, dfs):
        if not isinstance(dfs, (tuple, list)):
            raise TypeError("the input dataframes are not in a tuple or list")
        if not dfs:
            raise ValueError("no dataframes found")
        self._dataframes = tuple([ self._bear_child( SPTDataFrame, df ) for df in  dfs ])
        if not all([ tuple(self.columns) == tuple(df.columns) for df in dfs ]):
            raise ValueError("not all the dataframes feature the same column names")
    @property
    def reified(self):
        return True
    def __len__(self):
        return len(self._dataframes)
    def __iter__(self):
        yield from self._dataframes
    @property
    def columns(self):
        return self._dataframes[0].columns
    def reset_origin(self, columns=None, same_origin=False):
        if same_origin:
            if columns is None:
                columns = [ col for col in ['x', 'y', 'z', 't'] if col in self.columns ]
            origin = None
            for df in self.dataframes:
                _origin = df[columns].min().values
                if origin is None:
                    origin = _origin
                else:
                    origin = np.minimum(origin, _origin)
            for df in self.dataframes:
                df[columns] -= origin
        else:
            for df in self._dataframes:
                df.reset_origin(columns)

SPTData.register(SPTDataFrames)


class SPTFile(_SPTDataFrame):
    """
    Basis for :class:`SPTDataItem` classes for data in a file.
    """
    __slots__ = ('_alias','_reset_origin','_discard_static_trajectories')
    def __init__(self, filepath, dataframe=None, **kwargs):
        self._alias = None
        self._reset_origin = False
        self._discard_static_trajectories = False
        _SPTDataFrame.__init__(self, dataframe, filepath, **kwargs)
    @property
    def filepath(self):
        """
        *str*: Alias for the :attr:`source` property
        """
        return self._source
    @filepath.setter
    def filepath(self, fp):
        if self.reified:
            raise AttributeError("file '{}' has already been loaded; cannot set the filepath anymore".format(self.filepath.split('/')[-1]))
        else:
            self._source = fp
    @property
    def source(self):
        return self.filepath
    @source.setter
    def source(self, fp):
        self.filepath = fp
    @property
    def alias(self):
        """
        *str*: Identifier, shorter than :attr:`source`
        """
        return self._alias
    @alias.setter
    def alias(self, name):
        if callable(name):
            self._alias = name(self.filepath)
        else:
            self._alias = name
    def get_analyses(self):
        if self._analyses._data is None:
            self.load()
            assert self._analyses._data is not None or bool(self._analyses._instances)
        return self._analyses
    @property
    def analyses(self):
        return self.get_analyses()
    @analyses.setter
    def analyses(self, tree):
        self.set_analyses(tree)
    def _trigger_discard_static_trajectories(self):
        """
        Calls :meth:`SPTDataFrame.discard_static_trajectories`;
        requires the data are already loaded.
        """
        if self._discard_static_trajectories is True:
            SPTDataFrame.discard_static_trajectories(self)
        elif self._discard_static_trajectories: # not in (None, False)
            SPTDataFrame.discard_static_trajectories(self, **self._discard_static_trajectories)
    def _trigger_reset_origin(self):
        """
        Calls :meth:`SPTDataFrame.reset_origin`;
        requires the data are already loaded.
        """
        if self._reset_origin:
            SPTDataFrame.reset_origin(self, self._reset_origin)
    def reset_origin(self, columns=None, same_origin=False):
        if self.reified:
            SPTDataFrame.reset_origin(self, columns)
        else:
            self._reset_origin = True if columns is None else columns
    def discard_static_trajectories(self, dataframe=None, min_msd=None, **kwargs):
        if dataframe is not None or self.reified:
            return SPTDataFrame.discard_static_trajectories(self, dataframe, min_msd, **kwargs)
        else:
            if min_msd is not None:
                kwargs['min_msd'] = min_msd
            self._discard_static_trajectories = kwargs if kwargs else True
    def get_image(self, match=None):
        """
        Looks for the corresponding localization microscopy image in the
        :attr:`~tramway.analyzer.RWAnalyzer.images` attribute.

        The search is based on the object's alias.
        The first item that contains the alias in its filename is returned as a match.

        If the :attr:`alias` attribute is not set or images do not have a defined `filepath`,
        an :class:`AttributeError` exception is raised.

        The optional argument *match* is a 2-argument *callable* that
        takes the alias (*str*) of *self* and the filepath (*str*) of an
        :class:`~tramway.analyzer.images.Image` stack of images, and returns
        :const:`True` if the filepath and alias match.
        """
        if match is None:
            match = lambda alias, filepath: alias in os.path.basename(filepath)
        analyzer = self._eldest_parent
        if self.alias is None:
            try:
                return single(analyzer.images)
            except RuntimeError:
                raise AttributeError('alias is not defined') from None
        any_filepath_defined = False
        for image in analyzer.images:
            if image.filepath: # may raise AttributeError
                any_filepath_defined = True
                if match(self.alias, image.filepath):
                    return image
        if not any_filepath_defined:
            raise AttributeError('image filepaths are not defined')


class RawSPTFile(SPTFile):
    """
    Basis for :class:`SPTDataItem` classes for raw data in a file,
    possibly with non-standard column names or data units.
    """
    __slots__ = ('_columns',)
    def __init__(self, filepath, dataframe=None, **kwargs):
        self._columns = None
        SPTFile.__init__(self, filepath, dataframe, **kwargs)
    @property
    def columns(self):
        if self._columns is None:
            return self.dataframe.columns
        else:
            return self._columns
    @columns.setter
    def columns(self, cols):
        if self.reified:
            raise AttributeError('the SPT data have already been loaded; cannot set column names anymore')
        else:
            self._columns = cols

class _SPTAsciiFile(RawSPTFile):
    __slots__ = ()
    def load(self):
        self._dataframe = load_xyt(os.path.expanduser(self.filepath), self._columns, reset_origin=self._reset_origin)
        self._trigger_discard_static_trajectories()

class SPTAsciiFile(_SPTAsciiFile):
    __slots__ = ()

SPTDataItem.register(SPTAsciiFile)

class StandaloneSPTAsciiFile(_SPTAsciiFile, StandaloneDataItem):
    """
    `RWAnalyzer.spt_data` attribute for single SPT text files.
    """
    __slots__ = ()

SPTData.register(StandaloneSPTAsciiFile)


class SPTFiles(SPTDataFrames):
    __slots__ = ('_files','_filepattern')
    def __init__(self, filepattern, **kwargs):
        SPTDataIterator.__init__(self, **kwargs)
        self._files = []
        #self._filepattern = filepattern # should work in most cases
        if isinstance(filepattern, str):
            self._filepattern = os.path.expanduser(filepattern)
        else:
            self._filepattern = [os.path.expanduser(pattern) for pattern in filepattern]
    @property
    def filepattern(self):
        """
        *str*: Filepath glob pattern
        """
        return self._filepattern
    @filepattern.setter
    def filepattern(self, fp):
        if self._files:
            if fp != self._filepattern:
                raise AttributeError('the files have already been listed; cannot set the file pattern anymore')
        else:
            self._filepattern = fp
    @property
    def files(self):
        """
        *list*: SPT file objects
        """
        if not self._files:
            self.list_files()
        return self._files
    @property
    def filepaths(self):
        """
        *list* of *str*: File paths (copy)
        """
        return [ f.filepath for f in self.files ]
    @property
    def dataframes(self): # this also hides the parent class' setter
        return [ f.dataframe for f in self.files ]
    @property
    def partially_reified(self):
        """
        *bool*: :const:`True` if any file has been loaded
        """
        return self._files and any([ f.reified for f in self._files ])
    @property
    def fully_reified(self):
        """
        *bool*: :const:`True` if all the files have been loaded
        """
        return self._files and all([ f.reified for f in self._files ])
    @property
    def reified(self):
        return self.fully_reified
    def __len__(self):
        return len(self.files)
    def __iter__(self):
        yield from self.files
    def list_files(self):
        """
        Interprets the filepath glob pattern and lists the matching files.
        """
        if self.filepattern is None:
            raise ValueError('filepattern is not defined')
        if isinstance(self.filepattern, str):
            self._files = glob(self.filepattern)
        else:
            self._files = []
            for pattern in self.filepattern:
                self._files += glob(pattern)
        if not self._files:
            raise ValueError("no files found")
    @property
    def columns(self):
        it = iter(self)
        col = next(it).columns
        while True:
            try:
                _col = next(it).columns
            except StopIteration:
                break
            else:
                if _col != col:
                    raise AttributeError('not all the data blocks share the same columns')
        return col
    def reset_origin(self, columns=None, same_origin=False):
        if same_origin:
            SPTDataFrames.reset_origin(self, columns, True)
        else:
            for f in self.files:
                f.reset_origin(columns)
    @property
    def alias(self):
        """
        *callable*: Function that extracts an alias out of filepaths (**write-only property**)
        """
        raise AttributeError('alias is write-only')
    @alias.setter
    def alias(self, name):
        if callable(name):
            for f in self.files:
                f.alias = name
        else:
            raise TypeError('global alias is not callable')
    @property
    def aliases(self):
        """
        *list* of *str*: Aliases of the SPT data items (copy)
        """
        return [ f.alias for f in self.files ]
    def filter_by_source(self, source_filter, return_index=False):
        if return_index:
            _out = lambda i, f: i, f
        else:
            _out = lambda _, f: f
        # first test for aliases
        _any = False
        if source_filter is None:
            if return_index:
                yield from enumerate(self)
            else:
                yield from self
            _any = True
        elif callable(source_filter):
            for i, f in enumerate(self):
                if source_filter(f.alias):
                    yield _out(i, f)
                    _any = True
        else:
            if isinstance(source_filter, str):
                aliases = [source_filter]
            elif isinstance(source_filter, Sequence):
                yielded = set()
                for a in source_filter:
                    i, it = -1, iter(self)
                    try:
                        while True:
                            i += 1
                            f = next(it)
                            if f.alias == a:
                                if i in yielded:
                                    raise ValueError('duplicate source: {}'.format(a))
                                yield _out(i, f)
                                yielded.add(i)
                                _any = True
                                break
                    except StopIteration:
                        raise ValueError('cannot find source: {}'.format(a))
            elif isinstance(source_filter, Set):
                aliases = source_filter
            if not _any:
                for i, f in enumerate(self):
                    if f.alias in aliases:
                        yield _out(i, f)
                        _any = True
        # if no alias matched, try sources
        if not _any:
            yield from SPTDataIterator.filter_by_source(self, source_filter, return_index)

class RawSPTFiles(SPTFiles):
    __slots__ = ()
    @property
    def columns(self):
        return SPTFiles.columns.fget(self)
    @columns.setter
    def columns(self, col):
        for f in self:
            f.columns = col


class SPTAsciiFiles(RawSPTFiles):
    """
    :class:`SPTData` class for multiple SPT text files.
    """
    __slots__ = ()
    def list_files(self):
        SPTFiles.list_files(self)
        self._files = [ self._bear_child( SPTAsciiFile, filepath ) for filepath in self._files ]

SPTData.register(SPTAsciiFiles)


class _RWAFile(SPTFile):
    __slots__ = ('_filepath',)
    def __init__(self, filepath, **kwargs):
        self._filepath = filepath
        SPTFile.__init__(self, filepath, None, **kwargs)
    @property
    def filepath(self):
        """
        *str*: *.rwa* file path; distinct from :attr:`source`.
        """
        return self._filepath
    @filepath.setter
    def filepath(self, fp):
        if self.reified:
            raise AttributeError("file '{}' has already been loaded; cannot set the filepath anymore".format(self.filepath.split('/')[-1]))
        else:
            self._filepath = fp
    def set_analyses(self, tree):
        HasAnalysisTree.set_analyses(self, tree)
    def load(self):
        # ~ expansion is no longer necessary from rwa-python==0.8.4
        try:
            tree = load_rwa(os.path.expanduser(self.filepath),
                    lazy=True, force_load_spt_data=False)
            assert islazy(tree._data) or tree._data is None
            self.analyses = tree
        except KeyError as e:
            raise RWAFileException(self.filepath, e) from None
        self._trigger_discard_static_trajectories()
        self._trigger_reset_origin()
    @classmethod
    def __reload__(cls, self, parent=None, filepath=None):
        """
        To be called as either:

        .. code-block: python

            RWAFile.__reload__(spt_data_item)

        or:

        .. code-block: python

            StandaloneRWAFile.__reload__(spt_data)

        In the first case, the `parent` argument may be useful if the call to :meth:`__reload__`
        is part of a general reload of the :class:`SPTData` object which `spt_data_item` is an
        :class:`SPTDataItem` item.
        `parent` should point at the new reloaded :class:`SPTData` object.

        The returned object is of the most specialized type of `cls` and `type(self)`.

        """
        if filepath is None:
            if isinstance(self, _RWAFile):
                filepath = self.filepath
            else:
                filepath = os.path.splitext(self.source)[0]+'.rwa'
        if not os.path.isfile(os.path.expanduser(filepath)):
            raise FileNotFoundError(filepath)
        # favor the most specialized type
        if isinstance(self, cls):
            cls = type(self)
        #
        reloaded = cls(filepath, parent=self._parent if parent is None else parent)
        # should be readily available unless cache integrity has been compromised
        reloaded._frame_interval = self.frame_interval
        reloaded._localization_error = self.localization_error
        #
        reloaded._alias = self.alias
        #
        reloaded._roi = copy.copy(self.roi)
        reloaded.roi._parent = reloaded
        #
        reloaded._analyses.rwa_file = self._analyses.rwa_file
        reloaded._analyses.autosave = self._analyses.autosave
        self.close()
        return reloaded

class RWAFile(_RWAFile):
    __slots__ = ()

SPTDataItem.register(RWAFile)

class StandaloneRWAFile(_RWAFile, StandaloneDataItem):
    """
    :class:`SPTata` class for single RWA files.
    """
    __slots__ = ()
    def reload_from_rwa_files(self, skip_missing=False):
        if skip_missing:
            self.logger.error('cannot omit the only rwa file; ignoring skip_missing')
        cls = type(self) if isinstance(self, StandaloneRWAFile) else StandaloneRWAFile
        return self.self_update( cls.__reload__(self) )

SPTData.register(StandaloneRWAFile)


class RWAFiles(SPTFiles):
    """
    :class:`SPTData` class for multiple RWA files.
    """
    __slots__ = ()
    def list_files(self):
        SPTFiles.list_files(self)
        self._files = [ self._bear_child( RWAFile, filepath ) for filepath in self._files ]
    def reload_from_rwa_files(self, skip_missing=False):
        cls = type(self) if isinstance(self, RWAFiles) else RWAFiles
        assert cls is RWAFiles
        return self.self_update( cls.__reload__(self, skip_missing=skip_missing) )
    @classmethod
    def __reload__(cls, self, skip_missing=False, parent=None):
        if isinstance(self, cls):
            # favor the most specialized type
            cls = type(self)
        reloaded = cls([], parent=self._parent if parent is None else parent)
        reloaded.filepattern = None
        reloaded._files = []
        for f in self:
            try:
                f = RWAFile.__reload__(f, parent=reloaded)
            except FileNotFoundError:
                if not skip_missing:
                    raise
            else:
                reloaded._files.append(f)
        if not reloaded._files:
            raise RuntimeError('could not reload any .rwa file')
        return reloaded

SPTData.register(RWAFiles)


class _SPTMatFile(RawSPTFile):
    __slots__ = ('_coord_scale',)
    def __init__(self, filepath, dataframe=None, **kwargs):
        self._coord_scale = None
        RawSPTFile.__init__(self, filepath, dataframe, **kwargs)
    @property
    def coord_scale(self):
        r"""
        *float*:
            Convertion factor for the loaded coordinates so that they read in :math:`\mu m`
        """
        return self._coord_scale
    @coord_scale.setter
    def coord_scale(self, scale):
        if self.reified:
            raise AttributeError('the SPT data have already been loaded; cannot set the coordinate scale anymore')
        else:
            self._coord_scale = scale
    @property
    def pixel_size(self):
        """ *float*: Former name for :attr:`coord_scale`; **deprecated** """
        self.logger.warning('attribute pixel_size is deprecated; use coord_scale instead')
        return self.coord_scale
    @pixel_size.setter
    def pixel_size(self, siz):
        self.logger.warning('attribute pixel_size is deprecated; use coord_scale instead')
        self.coord_scale = siz
    def load(self):
        try:
            self._dataframe = load_mat(os.path.expanduser(self.filepath),
                    columns=self._columns, dt=self._frame_interval, coord_scale=self.coord_scale)
        except OSError as e:
            raise OSError('{}\nwhile loading file: {}'.format(e, self.source)) from None
        self._trigger_discard_static_trajectories()
        self._trigger_reset_origin()

class SPTMatFile(_SPTMatFile):
    __slots__ = ()

SPTDataItem.register(SPTMatFile)

class StandaloneSPTMatFile(_SPTMatFile, StandaloneDataItem):
    """
    :class:`SPTData` class for single MatLab v7 data files.
    """
    __slots__ = ()

SPTData.register(StandaloneSPTMatFile)


class SPTMatFiles(RawSPTFiles):
    """
    :class:`SPTData` class for multiple MatLab v7 data files.
    """
    __slots__ = ()
    @property
    def coord_scale(self):
        it = iter(self)
        px = next(it).coord_scale
        while True:
            try:
                _px = next(it).coord_scale
            except StopIteration:
                break
            else:
                _delta = px - _px
                if 1e-12 < _delta*_delta:
                    raise AttributeError('not all the data blocks share the same coordinate scale')
        return px
    @coord_scale.setter
    def coord_scale(self, scale):
        for f in self:
            f.coord_scale = scale
    coord_scale.__doc__ = SPTMatFile.coord_scale.__doc__
    @property
    def pixel_size(self):
        """ *float*: Former name for :attr:`coord_scale`; **deprecated** """
        self.logger.warning('attribute pixel_size is deprecated; use coord_scale instead')
        return self.coord_scale
    @pixel_size.setter
    def pixel_size(self, px):
        self.logger.warning('attribute pixel_size is deprecated; use coord_scale instead')
        self.coord_scale = px
    def list_files(self):
        SPTFiles.list_files(self)
        self._files = [ self._bear_child( SPTMatFile, filepath ) for filepath in self._files ]

SPTData.register(SPTMatFiles)


class RWGenerator(SPTDataFrame):
    """ not implemented yet """
    def __init__(self, *args, **kwargs):
        raise NotImplementedError


class RWAnalyses(StandaloneSPTDataFrame):
    """
    :class:`SPTData` class for single analysis trees as stored in *.rwa* files.

    .. note::

        copying (initializing with ``copy=True``) copies only the SPT data
        and NOT the subsequent analyses.

    .. warning::

        not thouroughly tested.

    """
    __slots__ = ()
    def __init__(self, analyses, copy=False, **kwargs):
        dataframe = analyses.data
        source = analyses.metadata.get('datafile', None)
        SPTDataFrame.__init__(self, dataframe, source, **kwargs)
        if not copy:
            self.analyses = analyses

class MultipleRWAnalyses(SPTDataFrames):
    """
    :class:`SPTData` class for multiple analysis trees as stored in *.rwa* files.

    .. note::

        copying (initializing with ``copy=True``) copies only the SPT data
        and NOT the subsequent analyses.

    .. warning::

        not tested.

    """
    __slots__ = ()
    def __init__(self, analyses, copy=False, **kwargs):
        dataframes = [ a.data for a in analyses ]
        SPTDataFrames.__init__(self, dataframes, **kwargs)
        if copy:
            for f, a in zip(self._dataframes, analyses):
                f.source = a.metadata.get('datafile', None)
        else:
            for f, a in zip(self._dataframes, analyses):
                f.analyses = a
        

class _FakeSPTData(pd.DataFrame):
    """
    Makes :class:`SPTDataFrames` initialization possible with empty data.

    To be wrapped in the :class:`SPTDataFrame` object for storage of SPT parameters.

    As soon as proper SPT datablocks are appended, the fake SPT datablock
    should be removed and the previously defined SPT parameters copied into
    the other datablocks.
    """
    __slots__ = () # useless as DataFrame defines __dict__
    def __init__(self, *args, **kwargs):
        pd.DataFrame.__init__(self, columns=list('nxyt'))

class LocalizationFile(SPTFile):
    __slots__ = ()

SPTDataItem.register(LocalizationFile)

class TrackerOutput(SPTDataFrames):
    """
    :class:`SPTData` class for SPT data yielded by the
    :attr:`~tramway.analyzer.RWAnalyzer.tracker` attribute.
    """
    __slots__ = ()
    def __init__(self, **kwargs):
        SPTDataFrames.__init__(self, [_FakeSPTData()], **kwargs)
    def add_tracked_data(self, trajectories, source=None, filepath=None):
        """
        .. note::

            In most cases `source` and `filepath` are supposed to represent
            the same piece of information.
            Here, `filepath` should be preferred over `source` if the localization
            data come from a file.

        """
        if filepath:
            df = self._bear_child( LocalizationFile, filepath, trajectories )
        else:
            df = self._bear_child( SPTDataFrame, trajectories, source )
        siblings = self._dataframes[0]
        df.frame_interval = siblings.frame_interval
        df.localization_error = siblings.localization_error
        df.temperature = siblings.temperature
        if isinstance(siblings.dataframe, _FakeSPTData):
            assert not self._dataframes[1:]
            self._dataframes = [ df ]
        else:
            if tuple(df.columns) != tuple(siblings.columns):
                raise ValueError("not all the dataframes feature the same column names")
            self._dataframes.append( df )
    # borrowed from `SPTFiles`
    @property
    def alias(self):
        raise AttributeError('alias is write-only')
    @alias.setter
    def alias(self, name):
        if callable(name):
            for f in self._dataframes:
                f.alias = name
        else:
            raise TypeError('global alias is not callable')
    @property
    def aliases(self):
        return [ f.alias for f in self._dataframes ]
    def filter_by_source(self, source_filter, return_index=False):
        yield from SPTFiles.filter_by_source(self, source_filter, return_index=return_index)


__all__ = [ 'SPTData', 'SPTDataItem', 'SPTParameters', 'StandaloneDataItem', 'SPTDataIterator',
        'SPTDataInitializer', 'HasROI', 'HasAnalysisTree', '_SPTDataFrame', 'SPTDataFrame',
        'StandaloneSPTDataFrame', 'SPTDataFrames',
        'SPTFile', 'RawSPTFile', 'SPTFiles', 'RawSPTFiles',
        '_SPTAsciiFile', 'SPTAsciiFile', 'StandaloneSPTAsciiFile', 'SPTAsciiFiles',
        '_RWAFile', 'RWAFile', 'StandaloneRWAFile', 'RWAFiles',
        '_SPTMatFile', 'SPTMatFile', 'StandaloneSPTMatFile', 'SPTMatFiles',
        'RWAnalyses', 'MultipleRWAnalyses', 'RWGenerator', 'LocalizationFile', 'TrackerOutput',
        'compute_dtypes', 'glob' ]




# -*- coding: utf-8 -*-

# Copyright © 2017-2020, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from math import tan, atan2, degrees, radians
import sys
import numpy as np
import pandas as pd
import numpy.ma as ma
from tramway.core.exceptions import NaNWarning
from tramway.tessellation import *
from tramway.inference import FiniteElements, Maps
from matplotlib.patches import Polygon, Wedge
from matplotlib.collections import PatchCollection
import scipy.spatial
import scipy.sparse as sparse
from warnings import warn


def _bounding_box(cells, xy=None):
    # bounding box
    try:
        bounding_box = cells.bounding_box[['x', 'y']]
        xy_min, xy_max = bounding_box.values
    except (KeyboardInterrupt, SystemExit):
        raise
    except:
        xy = cells.points[['x', 'y']]
        xy_min, xy_max = xy.min(axis=0), xy.max(axis=0)
    bounding_box = [ np.array(vs) for vs in (xy_min, [xy_min[0], xy_max[1]], xy_max, [xy_max[0], xy_min[1]]) ]
    return bounding_box


def cell_to_polygon(c, X, voronoi=None, bounding_box=None, region_point=None, return_voronoi=False):
    lstsq_kwargs = {}
    if sys.version_info[0] < 3:
        lstsq_kwargs['rcond'] = None
    if bounding_box is None:
        bounding_box = _bounding_box(None, X)
    if voronoi is None:
        voronoi = scipy.spatial.Voronoi(X)
    r = voronoi.point_region[c]
    region = voronoi.regions[r]
    if region_point is None:
        region_point = np.full(len(voronoi.regions), -1, dtype=int)
        region_point[voronoi.point_region] = np.arange(voronoi.point_region.size)
    #assert region_point[r] == c
    t = (bounding_box[0] + bounding_box[2]) * .5
    u = X[c]
    vertices = []
    for k, v in enumerate(region):
        if v < 0:

            # find the two "adjacent" vertices
            i, j = region[k-1], region[(k+1)%len(region)]
            assert 0<=i
            assert 0<=j
            # find the corresponding neighbour cells
            m, n = set(), set() # mutable
            for d, vs in enumerate(voronoi.regions):
                if not vs or d == r:
                    continue
                for e, cs in ((i,m), (j,n)):
                    try:
                        l = vs.index(e)
                    except ValueError:
                        continue
                    if (vs[l-1]==-1) or (vs[(l+1)%len(vs)]==-1):
                        cs.add(region_point[d])
            p, q = m.pop(), n.pop()
            assert not m
            assert not n
            # pick a distant point on the perpendicular bissector
            # of the neighbour centers, at the intersection with
            # the bounding box
            prev_edge = None
            for v, w in ((i,p), (j,q)):
                v, w = voronoi.vertices[v], X[w]
                if np.any(v<bounding_box[0]) or np.any(bounding_box[2]<v):
                    # vertex v stands outside the bounding box
                    continue
                n = np.array([u[1]-w[1], w[0]-u[0]])
                #w = (u + w) * .5
                w = v
                # determine direction: n or -n?
                if 0 < np.dot(n, t-w):
                    n = -n
                # intersection of [w,n) and [a,ab]
                z = None
                for l, a in enumerate(bounding_box):
                    b = bounding_box[(l+1)%len(bounding_box)]
                    M, p = np.c_[n, a-b], a-w
                    q = np.linalg.lstsq(M, p, **lstsq_kwargs)
                    q = q[0]
                    if 0<=q[0] and 0<=q[1] and q[1]<=1:
                        # intersection found
                        z = w + q[0] * n
                        #if 0<q[0]:
                        break
                assert z is not None
                if not (prev_edge is None or prev_edge == l):
                    # add the corner which the previous
                    # and current edges intersect at
                    e_max = len(bounding_box)-1
                    if (0<l and prev_edge==l-1) or \
                        (l==0 and prev_edge==e_max):
                        vertices.append(a)
                    elif (l<e_max and prev_edge==l+1) or \
                        (l==e_max and prev_edge==0):
                        vertices.append(b)
                    else:
                        raise RuntimeError
                prev_edge = l
                vertices.append(z)

        else:
            vertices.append(voronoi.vertices[v])
    if return_voronoi:
        return vertices, voronoi, bounding_box, region_point
    else:
        return vertices


def box_voronoi_2d(tessellation, xlim, ylim):
    not_a_vertex = -1
    points = tessellation.cell_centers
    vertices0 = tessellation.vertices
    cell_vertices0 = tessellation.cell_vertices
    Av0 = tessellation.vertex_adjacency.tocoo()
    xy0 = tessellation.cell_centers
    c0_inner = (xlim[0]<xy0[:,0]) & (ylim[0]<xy0[:,1]) & \
               (xy0[:,0]<xlim[1]) & (xy0[:,1]<ylim[1])
    c_inner_set = set(np.nonzero(c0_inner)[0])
    x, y = points[c0_inner,[0]], points[c0_inner,[1]]
    xy1 = np.r_[ xy0,
          np.c_[ 2*xlim[0]-x, y ],
          np.c_[ x, 2*ylim[0]-y ],
          np.c_[ 2*xlim[1]-x, y ],
          np.c_[ x, 2*ylim[1]-y ] ]
    c1_inner = np.r_[ c0_inner, np.zeros(4*len(x), dtype=c0_inner.dtype) ]
    voronoi1 = scipy.spatial.Voronoi(xy1)
    ridge_vertices1 = np.array(voronoi1.ridge_vertices)
    #
    r1_border = np.array([ c1_inner[r[0]] != c1_inner[r[1]] for r in voronoi1.ridge_points ])
    v1_border_set = set(ridge_vertices1[r1_border].flatten())
    assert -1 not in v1_border_set
    #
    v1_inner_set = set()
    for c in c_inner_set:
        vs = set(voronoi1.regions[voronoi1.point_region[c]])
        v1_inner_set |= set(vs)
    try:
        v1_inner_set.remove(-1)
    except KeyError:
        pass
    v1_inner_set -= v1_border_set
    #
    corners = np.array([ [ xlim[0], ylim[0] ],
                         [ xlim[0], ylim[1] ],
                         [ xlim[1], ylim[1] ],
                         [ xlim[1], ylim[0] ] ])
    # TODO: check corners are not in vertices0
    v1_inner = np.array(list(v1_inner_set | v1_border_set))
    v1_inner_map = np.full(len(voronoi1.vertices), not_a_vertex, dtype=int)
    v1_inner_map[v1_inner] = np.arange(len(v1_inner))
    V1 = voronoi1.vertices[v1_inner]
    V12 = .5 * np.sum(V1 * V1, axis=1, keepdims=True)
    D2 = np.dot(V1, corners.T) - V12 - .5* np.sum(corners*corners, axis=1, keepdims=True).T
    v1_corners = np.argmax(D2, axis=0)
    assert np.all(np.isclose(D2[v1_corners,np.arange(len(corners))], 0, atol=1e-6))
    v1_corners = v1_inner[v1_corners]
    v1_corners_set = set(v1_corners)
    assert v1_corners_set <= v1_border_set
    #
    v1_cntr = v1_cntr0 = len(vertices0)
    vertices1 = []
    cell_vertices1 = list(cell_vertices0)
    v01_new_edges, v11_new_edges = set(), set()
    v1_map = np.full(len(voronoi1.vertices), not_a_vertex, dtype=int)
    v0_keep = np.zeros(len(vertices0), dtype=bool)
    #
    v1_border = list(v1_border_set)
    vertices1.append(voronoi1.vertices[v1_border])
    v1_map[v1_border] = np.arange(v1_cntr, v1_cntr+len(v1_border))
    v1_cntr += len(v1_border)
    #
    for c in cell_vertices0:
        v0 = cell_vertices0[c]
        if c in c_inner_set:
            V0 = vertices0[v0]
            v1 = np.array(voronoi1.regions[voronoi1.point_region[c]])
            v1 = v1[0<=v1]
            V1 = voronoi1.vertices[v1]
            V02 = .5 * np.sum(V0 * V0, axis=1, keepdims=True)
            #V12_ = .5 * np.sum(V1 * V1, axis=1, keepdims=True)
            assert np.all(0<=v1_inner_map[v1])
            V12_ = V12[v1_inner_map[v1]]
            D2 = np.dot(V0, V1.T) - V02 - V12_.T
            v0c_match = np.argmax(D2, axis=0)
            v0c_keep = np.isclose(D2[v0c_match, np.arange(len(V1))], 0, atol=1e-6)
            if not np.any(v0c_keep):
                warn('the Voronoi diagrams totally mismatch at the border', RuntimeWarning)
                #continue
                #print(D2[v0c_match, np.arange(len(V1))])
                #import matplotlib.pyplot as plt
                #plt.scatter(V0[:,0], V0[:,1], 200, c='g', marker='+')
                #plt.scatter(V1[:,0], V1[:,1], 200, c='r', marker='x')
                ##plt.scatter(corners[:,0], corners[:,1], c='b', marker='o')
                #plt.show()
            v0_match = v0[v0c_match[v0c_keep]]
            v1_match = v1[v0c_keep]
            v1_replace = v1[~v0c_keep]
            v0_kept = np.unique(v0_match)
            v0_keep[v0_kept] = True
            #assert len(v0_kept) == np.sum(v0c_keep) # not sure about that
            v1_mapped = v1_map[v1_replace]
            v1_reused = v1_mapped[v1_mapped!=not_a_vertex]
            #
            V1_new = V1[~v0c_keep][v1_mapped==not_a_vertex]
            vertices1.append(V1_new)
            v1_new = np.arange(v1_cntr, v1_cntr+len(V1_new))
            v1_cntr += len(V1_new)
            #
            v1_map[v1_replace[v1_mapped==not_a_vertex]] = v1_new
            if v0_kept.size==0 and v1_new.size==0:
                cell_vertices1[c] = v1_reused
            else:
                cell_vertices1[c] = np.r_[np.array(list(v0_kept)), v1_reused, v1_new]
            #
            v1_set = set(v1)
            for _v0, _v1 in zip(v0_match, v1_match):
                _v1_neighbours = \
                        set(ridge_vertices1[ridge_vertices1[:,0]==_v1,1]) | \
                        set(ridge_vertices1[ridge_vertices1[:,1]==_v1,0])
                for _v1_neighbour in _v1_neighbours & v1_border_set:
                    v01_new_edges.add((_v0, _v1_neighbour))
                    #
                    if _v1_neighbour in v1_corners_set:
                        continue
                    _v1_neighbour_neighbours = \
                        set(ridge_vertices1[ridge_vertices1[:,0]==_v1_neighbour,1]) | \
                        set(ridge_vertices1[ridge_vertices1[:,1]==_v1_neighbour,0])
                    _corner = _v1_neighbour_neighbours & v1_corners_set
                    while _corner:
                        v11_new_edges.add((_v1_neighbour, _corner.pop()))
        else:
            v0_keep[v0] = True
    v0_map = np.full(len(vertices0), not_a_vertex, dtype=int)
    n0 = np.sum(v0_keep)
    v0_map[v0_keep] = np.arange(n0)
    assert n0 <= v1_cntr0
    v1_map[v1_map!=not_a_vertex] += n0 - v1_cntr0
    vertices1.insert(0, vertices0[v0_keep])
    vertices1 = np.vstack(vertices1)
    for c, _vs in enumerate(cell_vertices1):
        # some elements in cell_vertices1 are scalars; convert them into arrays
        if not isinstance(_vs, np.ndarray):
            _vs = cell_vertices1[c] = np.array([_vs])
        # for all cell at the border,
        if c in c_inner_set:
            # ...map the additional vertex indices so they match the compact vertices1
            _v1 = v1_cntr0 <= _vs
            _vs[_v1] += n0 - v1_cntr0
            if not np.all(_v1):
                # ...and similarly map the pre-existing vertices that are not discarded
                _vs[~_v1] = v0_map[_vs[~_v1]]
                assert not np.any(_vs == not_a_vertex)
    v01_new0, v01_new1 = zip(*v01_new_edges)
    v01_new0, v01_new1 = v0_map[np.array(v01_new0)], v1_map[np.array(v01_new1)]
    assert not np.any(v01_new0 == not_a_vertex)
    assert not np.any(v01_new1 == not_a_vertex)
    if v11_new_edges:
        v11_new0, v11_new1 = zip(*v11_new_edges)
        v11_new0, v11_new1 = v1_map[np.array(v11_new0)], v1_map[np.array(v11_new1)]
        assert not np.any(v11_new0 == not_a_vertex)
        assert not np.any(v11_new1 == not_a_vertex)
    else:
        v11_new0, v11_new1 = [], []
    Av1_rows, Av1_cols = v0_map[Av0.row], v0_map[Av0.col]
    _ok = (0<=Av1_rows) & (0<=Av1_cols)
    Av1_rows, Av1_cols = Av1_rows[_ok], Av1_cols[_ok]
    Av1 = sparse.csr_matrix((np.ones(len(Av1_rows)+2*len(v01_new_edges)+2*len(v11_new_edges), dtype=bool),
            (np.r_[Av1_rows, v01_new0, v01_new1, v11_new0, v11_new1],
             np.r_[Av1_cols, v01_new1, v01_new0, v11_new1, v11_new0])),
            shape=(len(vertices1), len(vertices1)))
    return vertices1, cell_vertices1, Av1


def scalar_map_2d(cells, values, aspect=None, clim=None, figure=None, axes=None, linewidth=1,
        delaunay=False, colorbar=True, alpha=None, colormap=None, unit=None, clabel=None,
        xlim=None, ylim=None, try_fix_corners=True, return_patches=False, **kwargs):
    """
    Plot a 2D scalar map as a colourful image.

    Arguments:

        cells (Partition or FiniteElements): spatial description of the cells

        values (pandas.DataFrame or numpy.ndarray): feature value at each cell,
            that will be represented as a colour

        aspect (str): passed to :func:`~matplotlib.axes.Axes.set_aspect`

        clim (2-element sequence): passed to :func:`~matplotlib.cm.ScalarMappable.set_clim`

        figure (matplotlib.figure.Figure): figure handle

        axes (matplotlib.axes.Axes): axes handle

        linewidth (int): cell border line width

        delaunay (bool or dict): overlay the Delaunay graph; if ``dict``, options are passed
            to :func:`~tramway.plot.mesh.plot_delaunay`

        colorbar (bool or str or dict): add a colour bar; if ``dict``, options are passed to
            :func:`~matplotlib.pyplot.colorbar`;
            setting colorbar to '*nice*' allows to produce a colorbar close to the figure
            of the same size as the figure

        unit/clabel (str): colorbar label, usually the unit of displayed feature

        alpha (float): alpha value of the cells

        colormap (str): colormap name; see also https://matplotlib.org/users/colormaps.html

        xlim (2-element sequence): lower and upper x-axis bounds

        ylim (2-element sequence): lower and upper y-axis bounds

        return_patches (bool): returns `PatchCollection` patches and the corresponding
            bin indices.

    Extra keyword arguments are passed to :func:`~matplotlib.collections.PatchCollection`.

    """
    coords = None
    if isinstance(values, pd.DataFrame):
        if values.shape[1] != 1:
            coords = values[[ col for col in 'xyzt' if col in values.columns ]]
            values = values[[ col for col in values.columns if col not in 'xyzt' ]]
            if values.shape[1] != 1:
                warn('multiple parameters available; mapping first one only', UserWarning)
        values = values.iloc[:,0] # to Series
    #values = pd.to_numeric(values, errors='coerce')

    # parse Delaunay-related arguments
    if delaunay:
        if not isinstance(delaunay, dict):
            delaunay = {}
        if linewidth and 'linewidth' not in delaunay:
            delaunay['linewidth'] = linewidth

    # turn the cells into polygons
    ids = []
    polygons = []
    if isinstance(cells, FiniteElements):

        ix, xy, ok = zip(*[ (i, c.center, bool(c)) for i, c in cells.items() ])
        ix, xy, ok = np.array(ix), np.array(xy), np.array(ok)
        if not (coords is None or np.all(np.isclose(xy, coords))): # debug
            print('coordinates mismatch')
            print('in map:')
            print(xy)
            print('in cells:')
            print(coords)
        rp = bb = voronoi = None
        for c in range(xy.shape[0]):
            if ok[c]:
                vertices, voronoi, bb, rp = cell_to_polygon(
                    c, xy, voronoi, bb, rp, True)
                polygons.append(Polygon(vertices, True))

    elif isinstance(cells, Partition) and isinstance(cells.tessellation, Voronoi):

        xy = cells.tessellation.cell_centers
        # copy/paste from below
        if not xlim or not ylim:
            xy_min, _, xy_max, _ = _bounding_box(cells, xy)
            if not xlim:
                xlim = (xy_min[0], xy_max[0])
            if not ylim:
                ylim = (xy_min[1], xy_max[1])
        ix = np.arange(xy.shape[0])
        from tramway.tessellation.hexagon import HexagonalMesh
        from tramway.tessellation.kdtree import KDTreeMesh
        if isinstance(cells.tessellation, (RegularMesh, HexagonalMesh, KDTreeMesh)):
            vertices, cell_vertices, Av = cells.tessellation.vertices, cells.tessellation.cell_vertices, cells.tessellation.vertex_adjacency.tocsr()
        else:
            try:
                vertices, cell_vertices, Av = box_voronoi_2d(cells.tessellation, xlim, ylim)
            except AssertionError:
                warn('could not fix the borders', RuntimeWarning)
                vertices, cell_vertices, Av = cells.tessellation.vertices, cells.tessellation.cell_vertices, cells.tessellation.vertex_adjacency.tocsr()
        try:
            ok = 0 < cells.location_count
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            print(traceback.format_exc())
            ok = np.ones(ix.size, dtype=bool)
        if cells.tessellation.cell_label is not None:
            ok = np.logical_and(ok, 0 < cells.tessellation.cell_label)
        map_defined = np.zeros_like(ok)
        map_defined[values.index] = True
        ok[np.logical_not(map_defined)] = False
        ok[ok] = np.logical_not(np.isnan(values.loc[ix[ok]].values))
        for i in ix[ok]:
            vs = cell_vertices[i].tolist()
            # order the vertices so that they draw a polygon
            v0 = v = vs[0]
            vs = set(vs)
            _vertices = []
            #vvs = [] # debug
            while True:
                _vertices.append(vertices[v])
                #vvs.append(v)
                vs.remove(v)
                if not vs:
                    break
                ws = set(Av.indices[Av.indptr[v]:Av.indptr[v+1]]) & vs
                if not ws:
                    ws = set(Av.indices[Av.indptr[v0]:Av.indptr[v0+1]]) & vs
                    if ws:
                        _vertices = _vertices[::-1]
                    else:
                        #print((v, vs, vvs, [Av.indices[Av.indptr[v]:Av.indptr[v+1]] for v in vs]))
                        warn('cannot find a path that connects all the vertices of a cell', RuntimeWarning)
                        if try_fix_corners:
                            vs = vertices[cell_vertices[i]]
                            if len(vs) != 3:
                                _min = vs.min(axis=0) == np.r_[xlim[0],ylim[0]]
                                _max = vs.max(axis=0) == np.r_[xlim[1],ylim[1]]
                                if _min[0]:
                                    vx = v0x = xlim[0]
                                    v1x = vs[:,0].max()
                                elif _max[0]:
                                    vx = v0x = xlim[1]
                                    v1x = vs[:,0].min()
                                else:
                                    vx = None
                                if _min[1]:
                                    vy = v1y = ylim[0]
                                    v0y = vs[:,1].max()
                                elif _max[1]:
                                    vy = v1y = ylim[1]
                                    v0y = vs[:,1].min()
                                else:
                                    vy = None
                                if vx is None or vy is None:
                                    vs = None
                                else:
                                    vs = np.array([[v0x,v0y],[vx,vy],[v1x,v1y]])
                            if vs is not None:
                                polygons.append(Polygon(vs, True))
                                ids.append(i)
                        break
                v = ws.pop()
            #
            if _vertices:
                _vertices = np.vstack(_vertices)
                polygons.append(Polygon(_vertices, True))
                ids.append(i)
    else:
        _type = repr(type(cells))
        if _type.endswith("'>"):
            _type = _type.split("'")[1]
        try:
            _nested_type = repr(type(cells.tessellation))
            if _nested_type.endswith("'>"):
                _nested_type = _nested_type.split("'")[1]
            raise TypeError('wrong type for `cells`: {}<{}>'.format(_type, _nested_type))
        except AttributeError:
            raise TypeError('wrong type for `cells`: {}'.format(_type))

    if not ids:
        ids = ix[ok]
    scalar_map = values.loc[ids].values

    #print(np.nonzero(~ok)[0])

    try:
        if np.any(np.isnan(scalar_map)):
            #print(np.nonzero(np.isnan(scalar_map)))
            msg = 'NaN found'
            try:
                warn(msg, NaNWarning)
            except:
                print('warning: {}'.format(msg))
            scalar_map[np.isnan(scalar_map)] = 0
    except TypeError: # help debug
        print(scalar_map)
        print(scalar_map.dtype)
        raise

    if figure is None:
        import matplotlib.pyplot as plt
        figure = plt.gcf() # before PatchCollection
    if axes is None:
        axes = figure.gca()
    # draw patches
    patch_kwargs = kwargs
    if alpha is not False:
        if alpha is None:
            alpha = .9
        patch_kwargs['alpha'] = alpha
    if colormap is not None:
        cmap = patch_kwargs.get('cmap', None)
        if cmap is None:
            patch_kwargs['cmap'] = colormap
        elif colormap != cmap:
            warn('both cmap and colormap arguments are passed with different values', RuntimeWarning)
    patches = PatchCollection(polygons, linewidth=linewidth, **patch_kwargs)
    patches.set_array(scalar_map)
    if clim is not None:
        patches.set_clim(clim)
    axes.add_collection(patches)

    obj = None
    if delaunay or isinstance(delaunay, dict):
        if not isinstance(delaunay, dict):
            delaunay = {}
        try:
            import tramway.plot.mesh as mesh
            obj = mesh.plot_delaunay(cells, centroid_style=None,
                axes=axes, **delaunay)
        except:
            import traceback
            traceback.print_exc()

    if not xlim or not ylim:
        xy_min, _, xy_max, _ = _bounding_box(cells, xy)
        if not xlim:
            xlim = (xy_min[0], xy_max[0])
        if not ylim:
            ylim = (xy_min[1], xy_max[1])
    axes.set_xlim(*xlim)
    axes.set_ylim(*ylim)
    if aspect is not None:
        axes.set_aspect(aspect)

    if colorbar:
        if colorbar=='nice':
            # make the colorbar closer to the plot and same size
            from mpl_toolkits.axes_grid1 import make_axes_locatable
            try:
                plt
            except NameError:
                import matplotlib.pyplot as plt
            try:
                gca_bkp = plt.gca()
                divider = make_axes_locatable(figure.gca())
                cax = divider.append_axes("right", size="5%", pad=0.05)
                _colorbar = figure.colorbar(patches, cax=cax)
                plt.sca(gca_bkp)
            except AttributeError as e:
                warn(e.args[0], RuntimeWarning)
        else:
            if not isinstance(colorbar, dict):
                colorbar = {}
            try:
                _colorbar = figure.colorbar(patches, ax=axes, **colorbar)
            except AttributeError as e:
                warn(e.args[0], RuntimeWarning)
        if clabel:
            unit = clabel
        if unit:
            _colorbar.set_label(unit)

    if return_patches:
        return patches, np.asarray(ids)
    else:
        return obj



def field_map_2d(cells, values, angular_width=30.0, overlay=False,
        aspect=None, figure=None, axes=None,
        cell_arrow_ratio=None,
        markercolor='y', markeredgecolor='k', markeralpha=0.8, markerlinewidth=None,
        transform=None,
        inferencemap=False, arrow_size=None, arrow_quantile_size=.1,
        **kwargs):
    """
    Plot a 2D field (vector) map as arrows.

    Arguments:

        cells (Partition or FiniteElements): spatial description of the cells

        values (pandas.DataFrame or numpy.ndarray): value at each cell, represented as a colour

        angular_width (float): angle of the tip of the arrows

        overlay (bool): do not plot the amplitude as a scalar map

        aspect (str): passed to :func:`~matplotlib.axes.Axes.set_aspect`

        figure (matplotlib.figure.Figure): figure handle

        axes (matplotlib.axes.Axes): axes handle

        cell_arrow_ratio (float): size of the largest arrow relative to the median
            inter-cell distance; default is ``0.4`` if `inferencemap` is ``True``,
            else ``1``

        markercolor (str): colour of the arrows

        markeredgecolor (str): colour of the border of the arrows

        markeralpha (float): alpha value of the arrows

        markerlinewidth (float): line width of the border of the arrows

        transform ('log', 'log10' or callable): if `overlay` is ``False``,
            transform applied to the amplitudes as a NumPy array;
            affects only the color-coded background map

        inferencemap (bool): if ``True``, the arrow length only depends on the cell size

        arrow_size (str): *new in 0.6*; can be :const:`'constant'`

        arrow_quantile_size (float): *new in 0.6*; applies if
            ``arrow_size=='constant'``; the base arrow size is set equal to a
            quantile of the inter-cell distance

    Extra keyword arguments are passed to :func:`scalar_map_2d` if called.

    If `overlay` is ``True``, the *marker*-prefixed arguments can be renamed without
    the *marker* prefix.
    These arguments can only be keyworded.

    """
    try:
        force_amplitude = values.pow(2).sum(1).apply(np.sqrt)
    except (KeyboardInterrupt, SystemExit):
        raise
    except:
        if not overlay:
            warn('cannot compute the amplitude; setting `overlay` to True', RuntimeWarning)
            overlay = True
    if figure is None:
        import matplotlib.pyplot as plt
        figure = plt.gcf()
    if axes is None:
        axes = figure.gca()
    #
    if transform is None:
        transform = lambda a: a
    elif transform == 'log':
        transform = np.log
    elif transform == 'log10':
        transform = np.log10
    #
    if overlay:
        if markercolor is None and 'color' in kwargs:
            markercolor = kwargs['color']
        if markeredgecolor is None and 'edgecolor' in kwargs:
            markeredgecolor = kwargs['edgecolor']
        if markeralpha is None and 'alpha' in kwargs:
            markeralpha = kwargs['alpha']
        if markerlinewidth is None and 'linewidth' in kwargs:
            markerlinewidth = kwargs['linewidth']
    else:
        obj = scalar_map_2d(cells, transform(force_amplitude),
            figure=figure, axes=axes, **kwargs)
    if aspect is not None:
        axes.set_aspect(aspect)
    xmin, xmax = axes.get_xlim()
    ymin, ymax = axes.get_ylim()
    if axes.get_aspect() == 'equal':
        aspect_ratio = 1
    else:
        aspect_ratio = (xmax - xmin) / (ymax - ymin)
    # identify the visible cell centers
    if isinstance(cells, Tessellation):
        pts = cells.cell_centers
    elif isinstance(cells, FiniteElements):
        pts = np.vstack([ cells[i].center for i in cells ])#values.index ])
    elif isinstance(cells, Partition):
        assert isinstance(cells.tessellation, Tessellation)
        pts = cells.tessellation.cell_centers#[values.index]
    inside = (xmin<=pts[:,0]) & (pts[:,0]<=xmax) & (ymin<=pts[:,1]) & (pts[:,1]<=ymax)
    # compute the distance between adjacent cell centers
    if isinstance(cells, Delaunay):
        A = cells.cell_adjacency
    elif isinstance(cells, FiniteElements):
        A = cells.adjacency
    elif isinstance(cells, Partition) and isinstance(cells.tessellation, Delaunay):
        A = cells.tessellation.cell_adjacency
    if inferencemap or arrow_size == 'constant':
        if cell_arrow_ratio is None:
            cell_arrow_ratio = 1.
        A = A.tocsr()
        scale = np.full(A.shape[0], np.nan, dtype=pts.dtype)
        for i in values.index:
            if not inside[i]:
                continue
            js = A.indices[A.indptr[i]:A.indptr[i+1]]
            pts_i, pts_j = pts[[i]], pts[js[inside[js]]]
            if pts_j.size:
                inter_cell_distance = pts_i - pts_j
                inter_cell_distance = np.sqrt(np.sum(inter_cell_distance * inter_cell_distance, axis=1))
                scale[i] = np.nanmean(inter_cell_distance)
        if arrow_size == 'constant':
            s = np.nanquantile(scale, arrow_quantile_size) / cell_arrow_ratio
            scale = np.full_like(scale, s)
        else:
            scale /= cell_arrow_ratio
    else:
        if cell_arrow_ratio is None:
            cell_arrow_ratio = .4 # backward compatibility
        A = sparse.triu(A, format='coo')
        I, J = A.row, A.col
        _inside = inside[I] & inside[J]
        pts_i, pts_j = pts[I[_inside]], pts[J[_inside]]
        inter_cell_distance = pts_i - pts_j
        inter_cell_distance = np.sqrt(np.sum(inter_cell_distance * inter_cell_distance, axis=1))
        # scale force amplitude
        large_arrow_length = np.nanmax(force_amplitude[inside[values.index]]) # TODO: clipping
        scale = np.nanmedian(inter_cell_distance) / (large_arrow_length * cell_arrow_ratio)
    #
    dw = float(angular_width) / 2.0
    t = tan(radians(dw))
    t = np.array([[0.0, -t], [t, 0.0]])
    markers = []
    for i in values.index:
        center = pts[i]
        radius = force_amplitude[i]
        if inferencemap or arrow_size == 'constant':
            f = np.asarray(values.loc[i])
            f *= scale[i] / max(1e-16, np.sqrt(np.sum(f * f)))
        else:
            f = np.asarray(values.loc[i]) * scale
        #fx, fy = f
        #angle = degrees(atan2(fy, fx))
        #markers.append(Wedge(center, radius, angle - dw, angle + dw))
        base, vertex = center + np.outer([-1./3, 2./3], f)
        ortho = np.dot(t, f)
        vertices = np.stack((vertex, base + ortho, base - ortho), axis=0)
        #vertices[:,0] = center[0] + aspect_ratio * (vertices[:,0] - center[0])
        markers.append(Polygon(vertices, True))

    patches = PatchCollection(markers, facecolor=markercolor, edgecolor=markeredgecolor,
        alpha=markeralpha, linewidth=markerlinewidth)
    axes.add_collection(patches)

    #axes.set_xlim(xmin, xmax)
    #axes.set_ylim(ymin, ymax)

    if not overlay and obj:
        return obj


def scalar_map_3d(cells, values, aspect=None, clim=None, figure=None, axes=None,
        colorbar=True, alpha=None, colormap=None, unit=None, clabel=None,
        xlim=None, ylim=None, zlim=None, triangulation_depth=2, **kwargs):
    """
    Plot a 2D scalar map as a colourful 3D surface.

    Arguments:

        cells (Tessellation or Partition): spatial description of the cells

        values (pandas.DataFrame, numpy.ndarray or Maps): feature value at each cell,
            that will be represented as a colour

        aspect (str): passed to :func:`~matplotlib.axes.Axes.set_aspect`

        clim (2-element sequence): passed to :func:`~matplotlib.cm.ScalarMappable.set_clim`;
            note `clim` affects colour, not height.

        figure (matplotlib.figure.Figure): figure handle

        axes (matplotlib.axes.Axes): axes handle

        colorbar (bool or str or dict): add a colour bar; if ``dict``, options are passed to
            :func:`~matplotlib.pyplot.colorbar`;
            setting colorbar to '*nice*' allows to produce a colorbar close to the figure
            of the same size as the figure

        unit/clabel (str): colorbar label, usually the unit of displayed feature

        alpha (float): alpha value of the cells

        colormap (str): colormap name; see also https://matplotlib.org/users/colormaps.html

        xlim (2-element sequence): lower and upper x-axis bounds

        ylim (2-element sequence): lower and upper y-axis bounds

        zlim (2-element sequence): lower and upper z-axis bounds;
            note `zlim` affects height, not colour.

    Extra keyword arguments are passed to :func:`~matplotlib.collections.PatchCollection`.

    """
    coords = None
    if isinstance(values, Maps):
        values = values.maps
    if isinstance(values, pd.DataFrame):
        if values.shape[1] != 1:
            coords = values[[ col for col in 'xyzt' if col in values.columns ]]
            values = values[[ col for col in values.columns if col not in 'xyzt' ]]
            if values.shape[1] != 1:
                warn('multiple parameters available; mapping first one only', UserWarning)
        values = values.iloc[:,0] # to Series

    try:
        mesh = cells.tessellation
    except AttributeError:
        mesh = cells
    try:
        mesh = mesh.spatial_mesh
    except AttributeError:
        pass
    xy = centers = mesh.cell_centers
    adjacency = mesh.simplified_adjacency(format='csr')

    scalar_map = np.full(centers.shape[0], np.nan)
    scalar_map[values.index] = values.values
    elevation = np.array(scalar_map)
    if zlim:
        elevation[elevation < zlim[0]] = zlim[0]
        elevation[zlim[1] < elevation] = zlim[1]
    colour = np.array(scalar_map)
    if clim:
        colour[colour < clim[0]] = clim[0]
        colour[clim[1] < colour] = clim[1]

    if figure is None:
        import matplotlib.pyplot as plt
        figure = plt.gcf() # before PatchCollection
    if axes is None:
        import mpl_toolkits.mplot3d as plt3
        axes = plt3.Axes3D(figure)

    obj = None

    if not xlim or not ylim:
        xy_min, _, xy_max, _ = _bounding_box(cells, xy)
        if not xlim:
            xlim = (xy_min[0], xy_max[0])
        if not ylim:
            ylim = (xy_min[1], xy_max[1])
    axes.set_xlim(*xlim)
    axes.set_ylim(*ylim)
    if aspect is not None:
        axes.set_aspect(aspect)

    tri = []
    ok = np.zeros(mesh.number_of_cells, dtype=bool)
    for i in range(mesh.number_of_cells):
        J = adjacency.indices[adjacency.indptr[i]:adjacency.indptr[i+1]]
        for j in J:
            if i < j:
                K = adjacency.indices[adjacency.indptr[j]:adjacency.indptr[j+1]]
                for k in K:
                    if j < k:
                        tri.append([i,j,k])
                        ok[i] = ok[j] = ok[k] = True
    tri = np.array(tri)
    #index_transform = np.full(mesh.number_of_cells, -1, dtype=int)
    #index_transform[ok] = np.arange(np.sum(ok))
    #import matplotlib.tri
    # if `plot_trisurf` could interpolate each triangle's colour:
    #tri_colour = [ np.mean(colour[_t]) for _t in tri ]
    #tri = matplotlib.tri.Triangulation(xy[ok,0], xy[ok,1], index_transform(tri))
    #axes.plot_trisurf(tri, elevation[ok], color=tri_colour, **kwargs)

    import itertools
    def split_triangle(_rdepth, _xy, *_zs):
        if _rdepth == 0:
            return tuple( [_a] for _a in (_xy,) + _zs )
        else:
            return tuple( itertools.chain(*_a) for _a in zip(
                    split_triangle(_rdepth-1,
                        np.stack((_xy[0], (_xy[0] + _xy[1]) * .5, (_xy[0] + _xy[2]) * .5), axis=0),
                        *[ np.r_[_z[0], (_z[0] + _z[1]) * .5, (_z[0] + _z[2]) * .5] for _z in _zs ]),
                    split_triangle(_rdepth-1,
                        np.stack(((_xy[0] + _xy[1]) * .5, _xy[1], (_xy[1] + _xy[2]) * .5), axis=0),
                        *[ np.r_[(_z[0] + _z[1]) * .5, _z[1], (_z[1] + _z[2]) * .5] for _z in _zs ]),
                    split_triangle(_rdepth-1,
                        np.stack(((_xy[0] + _xy[1]) * .5, (_xy[0] + _xy[2]) * .5, (_xy[1] + _xy[2]) * .5), axis=0),
                        *[ np.r_[(_z[0] + _z[1]) * .5, (_z[0] + _z[2]) * .5, (_z[1] + _z[2]) * .5] for _z in _zs ]),
                    split_triangle(_rdepth-1,
                        np.stack(((_xy[0] + _xy[2]) * .5, (_xy[1] + _xy[2]) * .5, _xy[2]), axis=0),
                        *[ np.r_[(_z[0] + _z[2]) * .5, (_z[1] + _z[2]) * .5, _z[2]] for _z in _zs ]),
                    ))

    subtri_xyz = []
    subtri_colour = []
    for ijk in tri:
        boundary_xy = xy[ijk]
        boundary_elevation = elevation[ijk]
        boundary_colour = colour[ijk]
        for _subtri_xy, _subtri_elevation, _subtri_colour in \
                zip(*split_triangle(triangulation_depth, boundary_xy, boundary_elevation, boundary_colour)):
            subtri_xyz.append(np.hstack((_subtri_xy, _subtri_elevation[:,np.newaxis])))
            subtri_colour.append(np.mean(_subtri_colour))
    subtri_xyz = np.vstack(subtri_xyz)
    #subtri = matplotlib.tri.Triangulation(
    #        subtri_xyz[:,0], subtri_xyz[:,1],
    #        np.reshape(np.arange(len(subtri_xyz)), (-1, 3), order='C'))
    #print(subtri.get_masked_triangles())
    subtri_xyz = np.stack([np.reshape(c, (-1,3)) for c in subtri_xyz.T], axis=-1)

    #obj = axes.plot_trisurf(subtri, subtri_elevation, facecolors=subtri_colour, norm=norm, **kwargs)
    import mpl_toolkits.mplot3d.art3d as art
    subtri = art.Poly3DCollection(subtri_xyz, **kwargs)
    subtri.set_array(np.asarray(subtri_colour))

    axes.add_collection(subtri)
    #axes.auto_scale_xyz(subtri_xyz[:,0], subtri_xyz[:,1], subtri_xyz[:,2])
    if xlim is not None:
        axes.set_xlim(xlim[0], xlim[1])
    if ylim is not None:
        axes.set_ylim(ylim[0], ylim[1])
    if zlim is not None:
        axes.set_zlim(zlim[0], zlim[1])

    return obj



__all__ = ['cell_to_polygon', 'scalar_map_2d', 'field_map_2d', 'scalar_map_3d']




# -*- coding: utf-8 -*-

# Copyright © 2020, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from ..attribute import *
from .abc import *
from . import environments


class EnvironmentInitializer(Initializer):
    """
    Can be left not-initialized.

    For local multi-processing:

    .. code:: python

        a.env = environments.LocalHost

    For *sbatch* job scheduling on an SSH-reachable server:

    .. code:: python

        a.env = environments.SlurmOverSSH

    See :class:`~.environments.LocalHost` and :class:`~.environments.SlurmOverSSH`.
    """
    __slots__ = ('_script',)
    def __init__(self, attribute_setter, parent=None):
        Initializer.__init__(self, attribute_setter, parent=parent)
        self._script = None
    def from_callable(self, env):
        if issubclass(env, Environment):
            self.specialize( env )
        else:
            raise TypeError('env is not an Environment')
    @property
    def script(self):
        """
        *str*: Path to the local script; this compatibility attribute
            is actually used by the :attr:`~tramway.analyzer.RWAnalyzer.browser`
        """
        #return None
        return self._script
    @script.setter
    def script(self, filename):
        #pass
        self._script = filename
    @property
    def collectibles(self):
        """ set: Compatilibity attribute; not used """
        return set()
    @property
    def submit_side(self):
        return False
    @property
    def worker_side(self):
        return False


__all__ = ['Environment', 'EnvironmentInitializer', 'environments']




# -*- coding: utf-8 -*-

# Copyright © 2020, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from ..abc import AnalyzerNode, TessellationPostProcessing
from tramway.helper.tessellation import Partition, delete_low_count_cells, update_cell_centers

class CellMerger(AnalyzerNode):
    __slots__ = ()
    def post_process(self, tessellation, spt_dataframe):
        # does nothing
        return tessellation

class ByTranslocationCount(CellMerger):
    """
    Merges the cells/microdomains that exhibit a number of translocations
    below some threshold.

    .. note::

        For now, all locations are counted instead of translocations only.

    """
    __slots__ = ('_count_threshold', '_update_centroids')
    def __init__(self, **kwargs):
        CellMerger.__init__(self, **kwargs)
        self._count_threshold = None
        self._update_centroids = None
    def post_process(self, tessellation, spt_dataframe):
        dim = tessellation.cell_centers.shape[1]
        min_ncells = dim + 2
        ncells = tessellation.number_of_cells
        cells = Partition(spt_dataframe, tessellation) # no options; Voronoi partition
        label = True
        while True:
            # TODO: consider translocations instead of any location
            cells, deleted_cells, label = delete_low_count_cells(
                    cells, self.count_threshold, 'count', label)
            if deleted_cells.size == 0:
                break
            if cells.number_of_cells < min_ncells:
                raise RuntimeError('too few remaining cells ({}<{})'.format(cells.number_of_cells, min_ncells))
            if self.update_centroids:
                cells = update_cell_centers(cells, self.update_centroids)
        return cells.tessellation
    @property
    def count_threshold(self):
        return self._count_threshold
    @count_threshold.setter
    def count_threshold(self, thres):
        self._count_threshold = thres
    @property
    def update_centroids(self):
        return self._update_centroids
    @update_centroids.setter
    def update_centroids(self, update):
        self._update_centroids = update

TessellationPostProcessing.register(ByTranslocationCount)


__all__ = ['ByTranslocationCount']

